@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/ftp.c 1.25 1998/12/03 19:21:51 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/ftp.c 1.33 1999/09/11 14:00:02 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.5.
   Functions for getting URLs using FTP.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1997,98 Andrew M. Bishop
+  This file Copyright 1997,98,99 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -21,6 +21,7 @@
 #include <sys/time.h>
 #include <unistd.h>
 #include <fcntl.h>
+#include <time.h>
 
 #include "wwwoffle.h"
 #include "misc.h"
@@ -97,7 +98,7 @@ char *FTP_Open(URL *Url)
 
  /* Open the connection. */
 
- server_ctrl=OpenClientSocket(server_host,server_port);
+ server_ctrl=OpenClientSocket(server_host,server_port,ConnectTimeout);
  init_buffer(server_ctrl);
 
  if(server_ctrl==-1)
@@ -116,35 +117,47 @@ char *FTP_Open(URL *Url)
 
   URL *Url The URL to get.
 
-  char *request_head The head of the HTTP request for the URL.
+  Header *request_head The head of the HTTP request for the URL.
 
-  char *request_body The body of the HTTP request for the URL.
+  Body *request_body The body of the HTTP request for the URL.
   ++++++++++++++++++++++++++++++++++++++*/
 
-char *FTP_Request(URL *Url,char *request_head,char *request_body)
+char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 {
  char *msg=NULL,*str=NULL;
  char *path,*file=NULL;
- char *host,*mimetype;
+ char *host,*mimetype=NULL;
  char *msg_reply=NULL;
+ char *timestamp,sizebuf[32];
  int i,l,port_l,port_h;
+ struct tm modtime;
+ time_t mtime;
  char *user,*pass;
 
+ /* Initial setting up. */
+
+ sizebuf[0]=0;
+ time(&mtime);
+ modtime=*gmtime(&mtime);
+
  /* Take a simple route if it is proxied. */
 
  if(proxy)
    {
-    char *new_request_head=MakeRequestAuthorised(proxy,request_head);
+    char *head;
 
-    PrintMessage(ExtraDebug,"Outgoing Request Head (to proxy)\n%s",new_request_head);
+    MakeRequestAuthorised(proxy,request_head);
 
-    if(write_string(server_ctrl,new_request_head)==-1)
+    head=HeaderString(request_head);
+
+    PrintMessage(ExtraDebug,"Outgoing Request Head (to proxy)\n%s",head);
+
+    if(write_string(server_ctrl,head)==-1)
        msg=PrintMessage(Warning,"Failed to write to remote FTP proxy [%!s].");
-    if(request_body && write_string(server_ctrl,request_body)==-1)
+    if(request_body && write_data(server_ctrl,request_body->content,request_body->length)==-1)
        msg=PrintMessage(Warning,"Failed to write to remote FTP proxy [%!s].");
 
-    if(request_head!=new_request_head)
-       free(new_request_head);
+    free(head);
 
     return(msg);
    }
@@ -180,10 +193,8 @@ char *FTP_Request(URL *Url,char *request_head,char *request_body)
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str);
-#if DEBUG_FTP
-    PrintMessage(Debug,"FTP: connected; got: %s",str);
-#endif
+    str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+    PrintMessage(ExtraDebug,"FTP: connected; got: %s",str);
 
     if(!file && !*path && str && isdigit(str[0]) && isdigit(str[1]) && isdigit(str[2]) && str[3]=='-')
       {
@@ -224,10 +235,8 @@ char *FTP_Request(URL *Url,char *request_head,char *request_body)
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str);
-#if DEBUG_FTP
-    PrintMessage(Debug,"FTP: sent 'USER %s'; got: %s",user,str);
-#endif
+    str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+    PrintMessage(ExtraDebug,"FTP: sent 'USER %s'; got: %s",user,str);
    }
  while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
@@ -254,10 +263,8 @@ char *FTP_Request(URL *Url,char *request_head,char *request_body)
 
     do
       {
-       str=read_line_or_timeout(server_ctrl,str);
-#if DEBUG_FTP
-       PrintMessage(Debug,"FTP: sent 'PASS %s'; got: %s",pass,str);
-#endif
+       str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+       PrintMessage(ExtraDebug,"FTP: sent 'PASS %s'; got: %s",pass,str);
 
        if(!file && !*path && str && isdigit(str[0]) && isdigit(str[1]) && isdigit(str[2]) && str[3]=='-')
          {
@@ -299,10 +306,8 @@ char *FTP_Request(URL *Url,char *request_head,char *request_body)
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str);
-#if DEBUG_FTP
-    PrintMessage(Debug,"FTP: sent 'CWD %s' got: %s",*path?path:"/",str);
-#endif
+    str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+    PrintMessage(ExtraDebug,"FTP: sent 'CWD %s' got: %s",*path?path:"/",str);
 
     if(!file && str && isdigit(str[0]) && isdigit(str[1]) && isdigit(str[2]) && str[3]=='-')
       {
@@ -345,10 +350,8 @@ char *FTP_Request(URL *Url,char *request_head,char *request_body)
 
     do
       {
-       str=read_line_or_timeout(server_ctrl,str);
-#if DEBUG_FTP
-       PrintMessage(Debug,"FTP: sent 'CWD %s' got: %s",file,str);
-#endif
+       str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+       PrintMessage(ExtraDebug,"FTP: sent 'CWD %s' got: %s",file,str);
       }
     while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
@@ -396,10 +399,8 @@ char *FTP_Request(URL *Url,char *request_head,char *request_body)
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str);
-#if DEBUG_FTP
-    PrintMessage(Debug,"FTP: sent 'TYPE %c'; got: %s",file?'I':'A',str);
-#endif
+    str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+    PrintMessage(ExtraDebug,"FTP: sent 'TYPE %c'; got: %s",file?'I':'A',str);
    }
  while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
@@ -416,6 +417,60 @@ char *FTP_Request(URL *Url,char *request_head,char *request_body)
     return(msg);
    }
 
+ if(!strcmp(request_head->method,"GET"))
+   {
+    /* Try and get the size and modification time. */
+
+    if(file)
+      {
+       if(write_formatted(server_ctrl,"SIZE %s\r\n",file)==-1)
+         {
+          msg=PrintMessage(Warning,"Failed to write 'SIZE' command to remote FTP host [%!s].");
+          return(msg);
+         }
+
+       do
+         {
+          str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+          PrintMessage(ExtraDebug,"FTP: sent 'SIZE %s'; got: %s",file,str);
+         }
+       while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
+
+       if (atoi(str) == 213)
+          if (str[4])
+             sprintf(sizebuf, "%ld", (long)atoi(str+4));
+
+       if(write_formatted(server_ctrl,"MDTM %s\r\n",file)==-1)
+         {
+          msg=PrintMessage(Warning,"Failed to write 'MDTM' command to remote FTP host [%!s].");
+          return(msg);
+         }
+
+       do
+         {
+          str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+          PrintMessage(ExtraDebug,"FTP: sent 'MDTM %s'; got: %s",file,str);
+         }
+       while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
+
+       if (atoi(str) == 213)
+         {
+          int year, mon, mday, hour, min, sec;
+
+          if (sscanf(str, "%*d %4d%2d%2d%2d%2d%2d", &year, &mon, &mday, &hour, &min, &sec) == 6)
+            {
+             memset(&modtime, 0, sizeof(modtime));
+             modtime.tm_year = year - 1900;
+             modtime.tm_mon = mon - 1;
+             modtime.tm_mday = mday;
+             modtime.tm_hour = hour;
+             modtime.tm_min = min;
+             modtime.tm_sec = sec;
+            }
+         }
+      }
+   }
+
  /* Create the data connection. */
 
  if(write_string(server_ctrl,"PASV\r\n")==-1)
@@ -426,10 +481,8 @@ char *FTP_Request(URL *Url,char *request_head,char *request_body)
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str);
-#if DEBUG_FTP
-    PrintMessage(Debug,"FTP: sent 'PASV'; got: %s",str);
-#endif
+    str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+    PrintMessage(ExtraDebug,"FTP: sent 'PASV'; got: %s",str);
    }
  while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
@@ -465,7 +518,7 @@ char *FTP_Request(URL *Url,char *request_head,char *request_body)
     if(host[l]==',')
        host[l]='.';
 
- server_data=OpenClientSocket(host,port_l+256*port_h);
+ server_data=OpenClientSocket(host,port_l+256*port_h,ConnectTimeout);
  init_buffer(server_data);
 
  if(server_data==-1)
@@ -476,89 +529,169 @@ char *FTP_Request(URL *Url,char *request_head,char *request_body)
 
  /* Make the request */
 
- if(file)
+ if(!strcmp(request_head->method,"GET"))
    {
-    if(write_formatted(server_ctrl,"RETR %s\r\n",file)==-1)
-       {
-        msg=PrintMessage(Warning,"Failed to write 'RETR' command to remote FTP host [%!s].");
-        return(msg);
-       }
-
-    do
+    if(file)
       {
-       str=read_line_or_timeout(server_ctrl,str);
-#if DEBUG_FTP
-       PrintMessage(Debug,"FTP: sent 'RETR %s'; got: %s",file,str);
-#endif
+       if(write_formatted(server_ctrl,"RETR %s\r\n",file)==-1)
+         {
+          msg=PrintMessage(Warning,"Failed to write 'RETR' command to remote FTP host [%!s].");
+          return(msg);
+         }
+
+       do
+         {
+          str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+          PrintMessage(ExtraDebug,"FTP: sent 'RETR %s'; got: %s",file,str);
+         }
+       while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
+
+       mimetype=WhatMIMEType(file);
       }
-    while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
-
-    mimetype=WhatMIMEType(file);
-   }
- else
-   {
-    if(write_string(server_ctrl,"LIST\r\n")==-1)
+    else
       {
-       msg=PrintMessage(Warning,"Failed to write 'LIST' command to remote FTP host [%!s].");
+       if(write_string(server_ctrl,"LIST\r\n")==-1)
+         {
+          msg=PrintMessage(Warning,"Failed to write 'LIST' command to remote FTP host [%!s].");
+          return(msg);
+         }
+
+       do
+         {
+          str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+          PrintMessage(ExtraDebug,"FTP: sent 'LIST'; got: %s",str);
+         }
+       while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
+
+       mimetype="text/html";
+      }
+
+    if(str && (atoi(str)==150 || atoi(str)==125))
+      {
+       char *p=str;
+       while(*p!='\r' && *p!='\n')
+          p++;
+       *p=0;
+      }
+    else
+      {
+       if(str)
+         {
+          char *p=str+strlen(str)-1;
+          while(*p=='\n' || *p=='\r') *p--=0;
+          msg=PrintMessage(Warning,"Got '%s' message after sending '%s' command to FTP server.",str,file?"RETR":"LIST");
+         }
+       else
+          msg=PrintMessage(Warning,"No reply from FTP server to '%s' command; timed out?",file?"RETR":"LIST");
+       return(msg);
+      }
+   }
+ else if(file) /* PUT */
+   {
+    if(file)
+      {
+       if(write_formatted(server_ctrl,"STOR %s\r\n",file)==-1)
+         {
+          msg=PrintMessage(Warning,"Failed to write 'STOR' command to remote FTP host [%!s].");
+          return(msg);
+         }
+
+       do
+         {
+          str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+          PrintMessage(ExtraDebug,"FTP: sent 'STOR %s'; got: %s",file,str);
+         }
+       while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
+
+       mimetype="text/html";
+      }
+    else
+      {
+       msg=PrintMessage(Warning,"Cannot use the PUT method on a directory name");
        return(msg);
       }
 
-    do
+    if(str && (atoi(str)==150 || atoi(str)==125))
       {
-       str=read_line_or_timeout(server_ctrl,str);
-#if DEBUG_FTP
-       PrintMessage(Debug,"FTP: sent 'LIST'; got: %s",str);
-#endif
-      }
-    while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
-
-    mimetype="text/html";
-   }
-
- if(str && (atoi(str)==150 || atoi(str)==125))
-   {
-    char *p=str;
-    while(*p!='\r' && *p!='\n')
-       p++;
-    *p=0;
-   }
- else
-   {
-    if(str)
-      {
-       char *p=str+strlen(str)-1;
-       while(*p=='\n' || *p=='\r') *p--=0;
-       msg=PrintMessage(Warning,"Got '%s' message after sending '%s' command to FTP server.",str,file?"RETR":"LIST");
+       char *p=str;
+       while(*p!='\r' && *p!='\n')
+          p++;
+       *p=0;
       }
     else
-       msg=PrintMessage(Warning,"No reply from FTP server to '%s' command; timed out?",file?"RETR":"LIST");
-    return(msg);
+      {
+       if(str)
+         {
+          char *p=str+strlen(str)-1;
+          while(*p=='\n' || *p=='\r') *p--=0;
+          msg=PrintMessage(Warning,"Got '%s' message after sending 'STOR' command to FTP server.",str);
+         }
+       else
+          msg=PrintMessage(Warning,"No reply from FTP server to 'STOR' command; timed out?");
+       return(msg);
+      }
+
+    write_data(server_data,request_body->content,request_body->length);
    }
 
  /* Prepare the HTTP header. */
 
- bufferhead=HTMLMessageHead(-1,200,str+4,
-                            "Content-Type",mimetype,
-                            NULL);
+ if(!strcmp(request_head->method,"GET"))
+   {
+    mtime=mktime(&modtime);
+    if(mtime!=-1)
+       timestamp=RFC822Date(mtime,1);
+    else
+       timestamp=RFC822Date(time(NULL),1);
+ 
+    if(*sizebuf)
+       bufferhead=HTMLMessageHead(-1,200,str+4,
+                                  "Content-Type",mimetype,
+                                  "Last-Modified",timestamp,
+                                  "Content-Length",sizebuf,
+                                  NULL);
+    else
+       bufferhead=HTMLMessageHead(-1,200,str+4,
+                                  "Content-Type",mimetype,
+                                  "Last-Modified",timestamp,
+                                  NULL);
+   }
+ else
+   {
+    bufferhead=HTMLMessageHead(-1,200,str+4,
+                               "Content-Type",mimetype,
+                               NULL);
+   }
+
  buffer=NULL;
  buffertail=NULL;
 
- if(!file)
+ if(!strcmp(request_head->method,"GET"))
    {
-    if(msg_reply)
+    if(!file)
       {
-       char *old_msg_reply=msg_reply;
-       msg_reply=HTMLString(msg_reply);
-       free(old_msg_reply);
-      }
+       if(msg_reply)
+         {
+          char *old_msg_reply=msg_reply;
+          msg_reply=HTMLString(msg_reply);
+          free(old_msg_reply);
+         }
 
-    buffer=HTMLMessageBody(-1,"FTPDir-Head",
+       buffer=HTMLMessageBody(-1,"FTPDir-Head",
+                              "url",URLDecode(Url->name,0),
+                              "message",msg_reply,
+                              NULL);
+
+       buffertail=HTMLMessageBody(-1,"FTPDir-Tail",
+                                  NULL);
+      }
+   }
+ else
+   {
+    buffer=HTMLMessageBody(-1,"FTPPut",
                            "url",URLDecode(Url->name,0),
-                           "message",msg_reply,
                            NULL);
 
-    buffertail=HTMLMessageBody(-1,"FTPDir-Tail",
-                               NULL);
    }
 
 near_end:
@@ -593,7 +726,7 @@ char *FTP_ReadHead(char *line)
  /* Take a simple route if it is proxied. */
 
  if(proxy)
-    return(read_line_or_timeout(server_ctrl,line));
+    return(read_line_or_timeout(server_ctrl,line,SocketTimeout));
  
  /* Else send the header. */
 
@@ -632,7 +765,7 @@ int FTP_ReadBody(char *s,int n)
  /* Take a simple route if it is proxied. */
 
  if(proxy)
-    return(read_data_or_timeout(server_ctrl,s,n));
+    return(read_data_or_timeout(server_ctrl,s,n,SocketTimeout));
  
  /* Else send the data then the tail. */
 
@@ -645,7 +778,7 @@ int FTP_ReadBody(char *s,int n)
        s[m]=buffertail[nreadtail];
    }
  else if(!buffer && !buffertail) /* File not dir entry. */
-    m=read_data_or_timeout(server_data,s,n);
+    m=read_data_or_timeout(server_data,s,n,SocketTimeout);
  else if(buffer && buffertail)  /* Middle of dir entry */
    {
     for(;nread<nbuffer && m<n;nread++,m++)
@@ -667,6 +800,11 @@ int FTP_ReadBody(char *s,int n)
     for(;nreadtail<nbuffertail && m<n;nreadtail++,m++)
        s[m]=buffertail[nreadtail];
    }
+ else /* if(buffer && !buffertail) */ /* Done a PUT */
+   {
+    for(;nread<nbuffer && m<n;nread++,m++)
+       s[m]=buffer[nread];
+   }
 
  return(m);
 }
@@ -690,24 +828,20 @@ int FTP_Close(void)
 
  /* Else say goodbye and close all of the sockets, */
 
+ if(server_data!=-1)
+    CloseSocket(server_data);
+
  write_string(server_ctrl,"QUIT\r\n");
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str);
-#if DEBUG_FTP
-    PrintMessage(Debug,"FTP: sent 'QUIT'; got: %s",str);
-#endif
+    str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+    PrintMessage(ExtraDebug,"FTP: sent 'QUIT'; got: %s",str);
    }
  while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
- /* Close the sockets */
-
  err=CloseSocket(server_ctrl);
 
- if(server_data!=-1)
-    CloseSocket(server_data);
-
  if(bufferhead)
     free(bufferhead);
 
@@ -734,7 +868,7 @@ static char *htmlise_dir_entry(char *line)
  int i,isdir=0;
  char *q,*p[16],*l,*file=NULL,*link=NULL;
 
- l=read_line_or_timeout(server_data,line);
+ l=read_line_or_timeout(server_data,line,SocketTimeout);
 
  if(!l)
     return(l);
