@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/io.c 2.13 1998/11/07 13:43:46 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/io.c 2.14 1999/07/03 10:48:22 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.5.
   Functions for file input and output.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97,98 Andrew M. Bishop
+  This file Copyright 1996,97,98,99 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -41,7 +41,6 @@
 #include <errno.h>
 
 #include "wwwoffle.h"
-#include "config.h"
 #include "misc.h"
 
 
@@ -58,7 +57,7 @@ static int nfdbuf=0;
 
 
 static int read_into_buffer(int fd);
-static int read_into_buffer_with_timeout(int fd);
+static int read_into_buffer_with_timeout(int fd,int timeout);
 static int read_from_buffer(int fd,char *buffer,int n);
 
 
@@ -202,9 +201,11 @@ int read_data(int fd,char *buffer,int n)
   char *buffer The buffer to put the data into.
 
   int n The number of bytes read.
+
+  int timeout The time in seconds to wait for the data.
   ++++++++++++++++++++++++++++++++++++++*/
 
-int read_data_or_timeout(int fd,char *buffer,int n)
+int read_data_or_timeout(int fd,char *buffer,int n,int timeout)
 {
  int nr;
 
@@ -219,7 +220,7 @@ int read_data_or_timeout(int fd,char *buffer,int n)
 
     FD_SET(fd,&readfd);
 
-    tv.tv_sec=SocketTimeout;
+    tv.tv_sec=timeout;
     tv.tv_usec=0;
 
     if(select(fd+1,&readfd,NULL,NULL,&tv)<=0)
@@ -293,10 +294,12 @@ char *read_line(int fd,char *line)
 
   char *line The previously allocated line of data.
 
+  int timeout The time in seconds to wait for the data.
+
   This is a replacement for the previous fgets_realloc() function with file descriptors instead of stdio.
   ++++++++++++++++++++++++++++++++++++++*/
 
-char *read_line_or_timeout(int fd,char *line)
+char *read_line_or_timeout(int fd,char *line,int timeout)
 {
  int found=0,eof=0;
  int n=0;
@@ -309,7 +312,7 @@ char *read_line_or_timeout(int fd,char *line)
     int i;
 
     if(!fdbytes[fd])
-       if(read_into_buffer_with_timeout(fd)<=0)
+       if(read_into_buffer_with_timeout(fd,timeout)<=0)
          {eof=1;break;}
 
     for(i=0;i<fdbytes[fd];i++)
@@ -361,9 +364,11 @@ static int read_into_buffer(int fd)
   int read_and_buffer Returns the number of bytes read.
 
   int fd The file descriptor to read from.
+
+  int timeout The time in seconds to wait for the data.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static int read_into_buffer_with_timeout(int fd)
+static int read_into_buffer_with_timeout(int fd,int timeout)
 {
  int n;
  fd_set readfd;
@@ -373,7 +378,7 @@ static int read_into_buffer_with_timeout(int fd)
 
  FD_SET(fd,&readfd);
 
- tv.tv_sec=SocketTimeout;
+ tv.tv_sec=timeout;
  tv.tv_usec=0;
 
  if(select(fd+1,&readfd,NULL,NULL,&tv)<=0)
