@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/misc.c 2.31 1998/11/30 19:16:58 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/misc.c 2.36 1999/07/02 19:37:57 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.5.
   Miscellaneous HTTP / HTML functions.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1997,98 Andrew M. Bishop
+  This file Copyright 1997,98,99 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -245,6 +245,19 @@ URL *SplitURL(char *url)
  else
     Url->link=Url->name;
 
+ Url->dir=Url->host;
+
+#if defined(__CYGWIN__)
+ if(strchr(Url->host,':'))
+   {
+    char *colon=strchr(Url->host,':');
+    Url->dir=(char*)malloc(strlen(Url->host));
+    strcpy(Url->dir,Url->host);
+
+    Url->dir[colon-Url->host]='!';
+   }
+#endif
+
  /* end */
 
  free(mallocurl);
@@ -310,6 +323,9 @@ void FreeURL(URL *Url)
  if(Url->name!=Url->file)
     free(Url->file);
 
+ if(Url->dir!=Url->host)
+    free(Url->dir);
+
  free(Url->name);
 
  if(Url->proto) free(Url->proto);
@@ -358,7 +374,6 @@ char *LinkURL(URL *Url,char *link)
       {
        int j;
        char *path=(char*)malloc(strlen(Url->path)+strlen(link)+2);
-       char *match;
 
        strcpy(path,Url->path);
 
@@ -370,36 +385,7 @@ char *LinkURL(URL *Url,char *link)
        strcat(path,"/");
        strcat(path,link);
 
-       while((match=strstr(path,"/../")))
-         {
-          char *prev=match; match+=3;
-          while(prev>path && *--prev!='/');
-          while((*prev++=*match++));
-         }
-
-       while((match=strstr(path,"/./")))
-         {
-          char *prev=match; match+=2;
-          while((*prev++=*match++));
-         }
-
-       while((match=strstr(path,"//")))
-         {
-          char *prev=match; match++;
-          while((*prev++=*match++));
-         }
-
-       match=&path[strlen(path)-2];
-       if(match>=path && !strcmp(match,"/."))
-          *match=0;
-
-       match=&path[strlen(path)-3];
-       if(match>=path && !strcmp(match,"/.."))
-          if(match==path)
-             *match=0;
-          else
-             while(match>path && *--match!='/')
-                *match=0;
+       CanonicaliseName(path);
 
        new=(char*)malloc(strlen(Url->proto)+strlen(Url->host)+strlen(path)+8);
        sprintf(new,"%s://%s%s",Url->proto,Url->host,path);
@@ -411,6 +397,76 @@ char *LinkURL(URL *Url,char *link)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Canonicalise a file name by removing '/../', '/./' and '//' references.
+
+  char *CanonicaliseName Returns the argument modified.
+
+  char *name The original name
+  ++++++++++++++++++++++++++++++++++++++*/
+
+char *CanonicaliseName(char *name)
+{
+ char *match,*name2;
+
+ match=name;
+ while((match=strstr(match,"/./")) || !strncmp(match=name,"./",2))
+   {
+    char *prev=match, *next=match+2;
+    while((*prev++=*next++));
+   }
+
+ match=name;
+ while((match=strstr(match,"//")))
+   {
+    char *prev=match, *next=match+1;
+    while((*prev++=*next++));
+   }
+
+ match=name2=name;
+ while((match=strstr(match,"/../")))
+   {
+    char *prev=match, *next=match+4;
+    if((prev-name2)==2 && !strncmp(name2,"../",3))
+      {name2+=3;match++;continue;}
+    while(prev>name2 && *--prev!='/');
+    match=prev;
+    if(*prev=='/')prev++;
+    while((*prev++=*next++));
+   }
+
+ match=&name[strlen(name)-2];
+ if(match>=name && !strcmp(match,"/."))
+    *match=0;
+
+ match=&name[strlen(name)-3];
+ if(match>=name && !strcmp(match,"/.."))
+    if(match==name)
+       *++match=0;
+    else
+       while(match>name && *--match!='/')
+          *match=0;
+
+#if 0 /* as used in cxref */
+
+ match=&name[strlen(name)-1];
+ if(match>name && !strcmp(match,"/"))
+    *match=0;
+
+ if(!*name)
+    *name='.',*(name+1)=0;
+
+#else /* as used in wwwoffle */
+
+ if(!*name || !strncmp(name,"../",3))
+    *name='/',*(name+1)=0;
+
+#endif
+
+ return(name);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Decode a string that has been UrlEncoded.
 
@@ -494,6 +550,48 @@ char *URLEncode(const char *str)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Encode a string using the UrlEncode method as used for passwords in URLs.
+
+  char *UrlEncodePassword Returns a malloced copy of the encoded string.
+
+  const char *str The string to be encoded.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+char *URLEncodePassword(const char *str)
+{
+ int i,j;
+ char *copy=(char*)malloc(3*strlen(str)+1);
+
+ for(i=0,j=0;str[i];i++)
+    if(isalpha(str[i]) ||
+       isdigit(str[i]) ||
+       str[i]=='?' || str[i]==':' || str[i]=='=' ||
+       str[i]=='$' || str[i]=='-' || str[i]=='_' || str[i]=='.' || str[i]=='+' || str[i]=='!' ||
+       str[i]=='*' || str[i]=='(' || str[i]==')' || str[i]==',' || str[i]=='\'')
+       copy[j++]=str[i];
+    else
+      {
+       unsigned int val=str[i];
+       copy[j]='%';
+       if(val%16>9)
+          copy[j+2]=val%16+'A'-10;
+       else
+          copy[j+2]=val%16+'0';
+       val/=16;
+       if(val%16>9)
+          copy[j+1]=val%16+'A'-10;
+       else
+          copy[j+1]=val%16+'0';
+       j+=3;
+      }
+
+ copy[j]=0;
+
+ return(copy);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Generate a hash value for a string.
 
