@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/purge.c 2.12 1998/11/30 20:20:39 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/purge.c 2.21 1999/08/18 18:46:30 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.5.
   Purge old files from the cache.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97 Andrew M. Bishop
+  This file Copyright 1996,97,98,99 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -25,6 +25,23 @@
 #include <dirent.h>
 #include <unistd.h>
 
+#if defined(__FreeBSD__) || defined(__NetBSD__)
+#include <sys/param.h>
+#include <sys/mount.h>
+#elif defined(__sun__) && defined(__svr4__)
+#include <sys/statvfs.h>
+#elif defined(__osf__)
+#include <sys/mount.h>
+#else
+#include <sys/vfs.h>
+#endif
+
+#if defined(__sun__) && defined(__svr4__)
+#define STATFS statvfs
+#else
+#define STATFS statfs
+#endif
+
 #include "wwwoffle.h"
 #include "misc.h"
 #include "proto.h"
@@ -33,7 +50,7 @@
 
 
 /* Local functions */
-static int PurgeFiles(char *proto,char *host);
+static long PurgeFiles(char *proto,char *host);
 
 /*+ Set this to 0 for debugging so that nothing is deleted. +*/
 #define DO_DELETE 1
@@ -41,14 +58,26 @@ static int PurgeFiles(char *proto,char *host);
 /*+ The current time. +*/
 static time_t now;
 
-/*+ The number of bytes left of each age. +*/
-static int *bytes_by_age;
+/*+ The number of blocks left of each age. +*/
+static int *blocks_by_age;
 
 /*+ The scaling factor for the ages in the second pass. +*/
 static double age_scale;
 
 /*+ The pass in the purge (1 or 2). +*/
-int pass;
+static int pass;
+
+/*+ The blocksize. +*/
+static long int blocksize;
+
+
+/*
+ Note: Since some UNIX versions (Solaris) have st_size/st_blocks different for
+       different files I have reverted to using 1+(st_size/blocksize) instead
+       of st_blocks for counting the number of blocks in a file.
+       This gives a small error in some cases but is fast to calculate compared
+       to the more accurate value of st_size/blocksize+!!(st_size%blocksize).
+*/
 
 
 /*++++++++++++++++++++++++++++++++++++++
@@ -59,36 +88,70 @@ int pass;
 
 void PurgeCache(int fd)
 {
- int total_bytes=0,total_dirs=0;
  int i,p;
- int age_for_size=-1;
+ long diskfree;
+ struct STATFS sbuf;
+ struct stat buf;
 
  now=time(NULL)+600;
 
- age_scale=0;
+ age_scale=-1;
 
- bytes_by_age=(int*)malloc((DefaultPurgeAge+1)*sizeof(int));
+ blocks_by_age=(int*)malloc((DefaultPurgeAge+2)*sizeof(int));
 
- for(i=0;i<=DefaultPurgeAge;i++)
-    bytes_by_age[i]=0;
+ for(i=0;i<=DefaultPurgeAge+1;i++)
+    blocks_by_age[i]=0;
 
- if(PurgeUseMTime)
-    write_string(fd,"(Using modification time.)\n");
+ if(stat(".",&buf) || buf.st_size==-1 || buf.st_blocks==-1)
+   {
+    PrintMessage(Warning,"Cannot determine the disk block size [%!s]; using 1024 instead.");
+    blocksize=1024;
+   }
  else
-    write_string(fd,"(Using last access time.)\n");
+   {
+    blocksize=buf.st_size/buf.st_blocks;
 
- if(PurgeUseURL)
-    write_string(fd,"(Using the full URL.)\n");
+    if(blocksize!=512 && blocksize!=1024 && blocksize!=2048)
+      {
+       PrintMessage(Warning,"The blocksize (%d) looks wrong; using 1024 instead.");
+       blocksize=1024;
+      }
+   }
+
+ if(STATFS(".",&sbuf) || sbuf.f_bsize==-1 || sbuf.f_bavail==-1)
+   {
+    PrintMessage(Warning,"Cannot determine the disk free space [%!s]; assuming 0.");
+    diskfree=0;
+   }
  else
-    write_string(fd,"(Using the hostname and protocol only.)\n");
+    diskfree=sbuf.f_bsize*sbuf.f_bavail/blocksize;
 
  for(pass=1;pass<=2;pass++)
    {
+    int total_blocks=0,total_dirs=0,dir_blocks;
+
+    write_string(fd,"\n");
+
     if(PurgeCacheSize)
        if(pass==1)
-          write_string(fd,"\nPass 1: Checking dates and sizes of files.\n");
+          write_string(fd,"Pass 1: Checking dates and sizes of files:\n");
        else
-          write_string(fd,"\nPass 2: Purging files down to specified size.\n");
+          write_string(fd,"Pass 2: Purging files down to specified size:\n");
+    else
+       write_string(fd,"Checking dates of files:\n");
+
+    if(pass==1)
+      {
+       if(PurgeUseMTime)
+          write_string(fd,"  (Using modification time.)\n");
+       else
+          write_string(fd,"  (Using last access time.)\n");
+
+       if(PurgeUseURL)
+          write_string(fd,"  (Using the full URL.)\n");
+       else
+          write_string(fd,"  (Using the hostname and protocol only.)\n");
+      }
 
     write_string(fd,"\n");
 
@@ -104,6 +167,11 @@ void PurgeCache(int fd)
        if(stat(proto,&buf))
          {PrintMessage(Inform,"Cannot stat directory '%s' [%!s]; not purged",proto);continue;}
 
+       dir_blocks=buf.st_size/blocksize;
+
+       total_blocks+=dir_blocks;
+       blocks_by_age[DefaultPurgeAge+1]+=dir_blocks;
+
        if(chdir(proto))
          {PrintMessage(Warning,"Cannot change to directory '%s' [%!s]; not purged.",proto);continue;}
 
@@ -123,12 +191,14 @@ void PurgeCache(int fd)
           struct stat buf;
 
           if(lstat(ent->d_name,&buf))
-             PrintMessage(Inform,"Cannot stat file '%s/%s' [%!s]; race condition?",proto,ent->d_name);
+             PrintMessage(Inform,"Cannot stat directory '%s/%s' [%!s]; race condition?",proto,ent->d_name);
           else if(S_ISDIR(buf.st_mode))
             {
-             int bytes=PurgeFiles(proto,ent->d_name);
+             long file_blocks=PurgeFiles(proto,ent->d_name);
 
-             if(bytes==-1)
+             dir_blocks=buf.st_size/blocksize;
+
+             if(file_blocks==-1)
                {
 #if DO_DELETE
                 if(rmdir(ent->d_name))
@@ -144,80 +214,117 @@ void PurgeCache(int fd)
                 utbuf.actime=buf.st_atime;
                 utbuf.modtime=buf.st_mtime;
                 utime(ent->d_name,&utbuf);
-               }
 
-             if(bytes>0)
-                total_bytes+=bytes;
-             if(bytes>=0)
+                blocks_by_age[DefaultPurgeAge+1]+=dir_blocks;
+                total_blocks+=file_blocks+dir_blocks;
                 total_dirs++;
+               }
 
              if(PurgeUseURL)
                {
-                if(bytes==-1)
-                   write_formatted(fd,"Purged %6s://%-32s ; (empty) - deleted.\n",proto,ent->d_name);
+                if(file_blocks==-1)
+                   write_formatted(fd,"Purged %6s://%-32s ; (empty) - deleted\n",proto,ent->d_name);
                 else
-                   write_formatted(fd,"Purged %6s://%-32s ; %5d kB.\n",proto,ent->d_name,(bytes+512)/1024);
+                   write_formatted(fd,"Purged %6s://%-32s ; %5ld kB\n",proto,ent->d_name,(file_blocks+dir_blocks)*blocksize/1024);
                }
              else
                {
-                int age=WhatPurgeAge(proto,ent->d_name,"/");
+                int age=WhatPurgeAge(proto,ent->d_name,"/",NULL);
 
                 if(pass==2 && age>0)
-                   age=(int)(age*age_scale);
+                   age=(int)(age*age_scale+0.5);
 
                 if(age<0)
-                   write_formatted(fd,"Not Purged       %6s://%-32s ; %5d kB.\n",proto,ent->d_name,(bytes+512)/1024);
-                else if(bytes==-1)
-                   write_formatted(fd,"Purged (%2d days) %6s://%-32s ; (empty) - deleted.\n",age,proto,ent->d_name);
+                   write_formatted(fd,"Not Purged       %6s://%-32s ; %5ld kB\n",proto,ent->d_name,(file_blocks+dir_blocks)*blocksize/1024);
+                else if(file_blocks==-1)
+                   write_formatted(fd,"Purged (%2d days) %6s://%-32s ; (empty) - deleted\n",age,proto,ent->d_name);
                 else
-                   write_formatted(fd,"Purged (%2d days) %6s://%-32s ; %5d kB.\n",age,proto,ent->d_name,(bytes+512)/1024);
+                   write_formatted(fd,"Purged (%2d days) %6s://%-32s ; %5ld kB\n",age,proto,ent->d_name,(file_blocks+dir_blocks)*blocksize/1024);
                }
             }
+          else
+            {
+             PrintMessage(Warning,"Found an unexpected file instead of a directory '%s/%s' [%!s]; deleting it.",proto,ent->d_name);
+
+#if DO_DELETE
+             if(unlink(ent->d_name))
+                PrintMessage(Warning,"Cannot delete the non-directory '%s/%s' [%!s].",proto,ent->d_name);
+#else
+             PrintMessage(Debug,"unlink(%s/%s).",proto,ent->d_name);
+#endif
+            }
+
          }
 
        closedir(dir);
        chdir("..");
       }
 
-    write_formatted(fd,"Total of %d directories ; %d kB.\n",total_dirs,(total_bytes+512)/1024);
+    write_string(fd,"\n");
+    write_formatted(fd,"Total of %d directories ; %ld kB\n",total_dirs,total_blocks*blocksize/1024);
 
     if(pass==1)
       {
+       int age_for_size=-1,age_for_free=-1;
+       int age_blocks_used=blocks_by_age[DefaultPurgeAge+1];
+       int age_blocks_free=diskfree+total_blocks-blocks_by_age[DefaultPurgeAge+1];
+
        write_string(fd,"\n");
-       write_string(fd,"(All ages scaled to the range 0 -> default age.)\n");
+       write_string(fd,"Age Profile of cached pages:\n");
+       write_string(fd,"  (All ages scaled to the range 0 -> default age.)\n");
        write_string(fd,"\n");
 
-       for(total_bytes=i=0;i<=DefaultPurgeAge;i++)
+       write_formatted(fd,"Total not purged   ; %5ld kB (%6ld kB free)\n",
+                       age_blocks_used*blocksize/1024,age_blocks_free*blocksize/1024);
+       write_string(fd,"\n");
+
+       for(i=0;i<=DefaultPurgeAge;i++)
          {
-          total_bytes+=bytes_by_age[i];
+          age_blocks_used+=blocks_by_age[i];
+          age_blocks_free-=blocks_by_age[i];
 
-          if(PurgeCacheSize && age_for_size<0 && ((total_bytes+512)/1024)>(1024*PurgeCacheSize))
+          if(PurgeCacheSize && age_for_size<0 && (age_blocks_used*blocksize/1024)>(1024*PurgeCacheSize))
             {
-             if(DefaultPurgeAge)
-               {
-                age_for_size=i;
-                age_scale=(double)age_for_size/(double)DefaultPurgeAge;
-               }
-             else
-                age_scale=0;
-             write_formatted(fd,"Cutoff Age is %2d days for %3d MB.\n",age_for_size,PurgeCacheSize);
+             age_for_size=i;
+
+             write_formatted(fd,"Cutoff Age is %2d days for %3d MB cache size\n",age_for_size,PurgeCacheSize);
+            }
+
+          if(PurgeDiskFree && diskfree && age_for_free<0 && (age_blocks_free*blocksize/1024)<(1024*PurgeDiskFree))
+            {
+             age_for_free=i;
+
+             write_formatted(fd,"Cutoff Age is %2d days for %3d MB disk free\n",age_for_free,PurgeDiskFree);
             }
 
           if(i==DefaultPurgeAge)
-             write_formatted(fd,"Total all ages     ; %5d kB.\n",(total_bytes+512)/1024);
+             write_formatted(fd,"Total all ages     ; %5ld kB (%6ld kB free)\n",
+                             (age_blocks_used-blocks_by_age[DefaultPurgeAge+1])*blocksize/1024,
+                             diskfree?age_blocks_free*blocksize/1024:0);
           else
-             write_formatted(fd,"Newer than %2d day%c ; %5d kB.\n",i+1,i?'s':' ',(total_bytes+512)/1024);
+             write_formatted(fd,"Newer than %2d day%c ; %5ld kB (%6ld kB free)\n",i+1,i?'s':' ',
+                             (age_blocks_used-blocks_by_age[DefaultPurgeAge+1])*blocksize/1024,
+                             diskfree?age_blocks_free*blocksize/1024:0);
          }
-       total_bytes=0;
+
+       if(DefaultPurgeAge)
+         {
+          if(age_for_size!=-1 && (age_for_size<=age_for_free || age_for_free==-1))
+             age_scale=(double)age_for_size/(double)DefaultPurgeAge;
+          else if(age_for_free!=-1 && (age_for_free<age_for_size || age_for_size==-1))
+             age_scale=(double)age_for_free/(double)DefaultPurgeAge;
+         }
+       else if(age_for_size!=-1 || age_for_free!=-1)
+          age_scale=0;
       }
 
-    if(!PurgeCacheSize || age_for_size==-1)
+    if(age_scale==-1)
        break;
    }
 
  write_string(fd,"\n");
 
- free(bytes_by_age);
+ free(blocks_by_age);
 
  /* Purge the tmp.* files in outgoing. */
 
@@ -262,25 +369,69 @@ void PurgeCache(int fd)
    }
 
  chdir("..");
+
+ /* Purge the tmp.* files in temp. */
+
+ if(chdir("temp"))
+    PrintMessage(Warning,"Cannot change to directory 'temp' [%!s]; not purged.");
+ else
+   {
+    DIR *dir;
+    struct dirent* ent;
+
+    dir=opendir(".");
+    if(!dir)
+      PrintMessage(Warning,"Cannot open directory 'temp' [%!s]; not purged.");
+    else
+      {
+       ent=readdir(dir);  /* skip .  */
+       if(!ent)
+          PrintMessage(Warning,"Cannot read directory 'temp' [%!s]; not purged.");
+       else
+         {
+          ent=readdir(dir);  /* skip .. */
+
+          while((ent=readdir(dir)))
+             if(!strncmp(ent->d_name,"tmp.",4))
+               {
+                struct stat buf;
+
+                if(!stat(ent->d_name,&buf) && buf.st_mtime<(now-60))
+                  {
+#if DO_DELETE
+                   if(unlink(ent->d_name))
+                      PrintMessage(Warning,"Cannot unlink file 'temp/%s' [%!s].",ent->d_name);
+#else
+                   PrintMessage(Debug,"unlink(temp/%s).",ent->d_name);
+#endif
+                  }
+               }
+         }
+
+       closedir(dir);
+      }
+   }
+
+ chdir("..");
 }
 
 
 /*++++++++++++++++++++++++++++++++++++++
   Delete the file in the current directory that are older than the specified age.
 
-  int PurgeFiles Returns the number of bytes in files that are left.
+  long PurgeFiles Returns the number of blocks in files that are left.
 
   char *proto The name of the protocol directory to purge.
 
   char *host The name of the host directory to purge.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static int PurgeFiles(char *proto,char *host)
+static long PurgeFiles(char *proto,char *host)
 {
- int bytes_left=-1;
+ long blocks_left=-1;
  DIR *dir;
  struct dirent* ent;
- int def_age=WhatPurgeAge(proto,host,"/");
+ int def_age=WhatPurgeAge(proto,host,"/",NULL);
 
  /* Open the spool subdirectory. */
 
@@ -303,7 +454,10 @@ static int PurgeFiles(char *proto,char *host)
     struct stat buf,buf2;
 
     if(stat(ent->d_name,&buf))
+       ;
+/*
        PrintMessage(Inform,"Cannot stat file '%s/%s/%s' [%!s]; race condition?",proto,host,ent->d_name);
+*/
     else
       {
        int age=DefaultPurgeAge;
@@ -340,7 +494,7 @@ static int PurgeFiles(char *proto,char *host)
              if(url)
                {
                 URL *Url=SplitURL(url);
-                age=WhatPurgeAge(proto,host,Url->path);
+                age=WhatPurgeAge(Url->proto,Url->host,Url->path,Url->args);
                 FreeURL(Url);
                 free(url);
                }
@@ -362,15 +516,15 @@ static int PurgeFiles(char *proto,char *host)
          }
 
        if(pass==2 && age>0)
-          age=(int)(age*age_scale);
+          age=(int)(age*age_scale+0.5);
 
        if(age==-1 || t>(now-age*(24*3600)))
          {
-          int size=buf.st_size+buf2.st_size;
+          long size=2+(buf.st_size+buf2.st_size)/blocksize;
 
-          if(bytes_left==-1)
-             bytes_left=0;
-          bytes_left+=size;
+          if(blocks_left==-1)
+             blocks_left=0;
+          blocks_left+=size;
 
           if(age>0)
             {
@@ -380,8 +534,10 @@ static int PurgeFiles(char *proto,char *host)
 
              if(days>DefaultPurgeAge)
                 days=DefaultPurgeAge;
-             bytes_by_age[days]+=size;
+             blocks_by_age[days]+=size;
             }
+          else
+             blocks_by_age[DefaultPurgeAge+1]+=size;
          }
        else
          {
@@ -410,5 +566,5 @@ static int PurgeFiles(char *proto,char *host)
  closedir(dir);
  chdir("..");
 
- return(bytes_left);
+ return(blocks_left);
 }
