@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/spool.c 2.37 1998/11/22 20:36:07 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/spool.c 2.49 1999/10/19 16:38:11 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.5b.
   Handle all of the spooling of files in the spool directory.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97,98 Andrew M. Bishop
+  This file Copyright 1996,97,98,99 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -32,6 +32,12 @@
 #include "config.h"
 
 
+/*+ Need this for Win32 to use binary mode +*/
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Open a file in the outgoing directory to write into / read from.
 
@@ -53,7 +59,7 @@ int OpenOutgoingSpoolFile(int rw)
  if(lstat("outgoing",&buf))
    {
     PrintMessage(Inform,"Directory 'outgoing' does not exist [%!s]; creating one.");
-    if(mkdir("outgoing",0755))
+    if(mkdir("outgoing",DirPerm))
       {PrintMessage(Warning,"Cannot create directory 'outgoing' [%!s].");return(-1);}
    }
  else
@@ -85,7 +91,9 @@ int OpenOutgoingSpoolFile(int rw)
              PrintMessage(Inform,"Cannot rename file 'outgoing/%s' to 'outgoing/%s' [%!s]; race condition?",ent->d_name,name);
           else
             {
-             if((fd=open(name,O_RDONLY))==-1)
+             fd=open(name,O_RDONLY|O_BINARY);
+
+             if(fd==-1)
                 PrintMessage(Inform,"Cannot open file 'outgoing/%s' [%!s]; race condition?",name);
              else
                {
@@ -100,7 +108,7 @@ int OpenOutgoingSpoolFile(int rw)
    }
  else
    {
-    fd=open(name,O_WRONLY|O_CREAT|O_EXCL,0644);
+    fd=open(name,O_WRONLY|O_CREAT|O_EXCL|O_BINARY,FilePerm);
     if(fd==-1)
        PrintMessage(Warning,"Cannot open file 'outgoing/%s' [%!s]",name);
    }
@@ -141,7 +149,7 @@ void CloseOutgoingSpoolFile(int fd,URL *Url)
  *newname='U';
 
  unlink(newname);
- ufd=open(newname,O_WRONLY|O_CREAT|O_EXCL,0644);
+ ufd=open(newname,O_WRONLY|O_CREAT|O_EXCL|O_BINARY,FilePerm);
  if(ufd!=-1)
    {
     write_string(ufd,Url->file);
@@ -215,7 +223,7 @@ char *HashOutgoingSpoolFile(URL *Url)
  name=URLToFileName(Url);
  *name='O';
 
- fd=open(name,O_RDONLY);
+ fd=open(name,O_RDONLY|O_BINARY);
  if(fd==-1)
    {PrintMessage(Warning,"Cannot open outgoing request to create hash [%!s].");free(name);chdir("..");return(NULL);}
 
@@ -337,7 +345,7 @@ int OpenWebpageSpoolFile(int rw,URL *Url)
  if(lstat(Url->proto,&buf))
    {
     PrintMessage(Inform,"Directory '%s' does not exist [%!s]; creating one.",Url->proto);
-    if(mkdir(Url->proto,0755))
+    if(mkdir(Url->proto,DirPerm))
       {PrintMessage(Warning,"Cannot create directory '%s' [%!s].",Url->proto);return(-1);}
    }
  else
@@ -347,18 +355,18 @@ int OpenWebpageSpoolFile(int rw,URL *Url)
  if(chdir(Url->proto))
    {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return(-1);}
 
- if(lstat(Url->host,&buf))
+ if(lstat(Url->dir,&buf))
    {
-    PrintMessage(Inform,"Directory '%s/%s' does not exist [%!s]; creating one.",Url->proto,Url->host);
-    if(mkdir(Url->host,0755))
-      {PrintMessage(Warning,"Cannot create directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return(-1);}
+    PrintMessage(Inform,"Directory '%s/%s' does not exist [%!s]; creating one.",Url->proto,Url->dir);
+    if(mkdir(Url->dir,DirPerm))
+      {PrintMessage(Warning,"Cannot create directory '%s/%s' [%!s].",Url->proto,Url->dir);chdir("..");return(-1);}
    }
  else
     if(!S_ISDIR(buf.st_mode))
-      {PrintMessage(Warning,"The file '%s/%s' is not a directory.",Url->proto,Url->host);chdir("..");return(-1);}
+      {PrintMessage(Warning,"The file '%s/%s' is not a directory.",Url->proto,Url->dir);chdir("..");return(-1);}
 
- if(chdir(Url->host))
-   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return(-1);}
+ if(chdir(Url->dir))
+   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);chdir("..");return(-1);}
 
  /* Open the file for the web page. */
 
@@ -366,16 +374,16 @@ int OpenWebpageSpoolFile(int rw,URL *Url)
 
  *file='D';
  if(rw)
-    fd=open(file,O_RDONLY);
+    fd=open(file,O_RDONLY|O_BINARY);
  else
-    fd=open(file,O_RDWR|O_CREAT,0644);
+    fd=open(file,O_RDWR|O_CREAT|O_BINARY,FilePerm);
 
  if(!rw && fd!=-1)
    {
     int ufd;
 
     *file='U';
-    ufd=open(file,O_WRONLY|O_CREAT|O_TRUNC,0644);
+    ufd=open(file,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,FilePerm);
 
     if(ufd!=-1)
       {
@@ -428,8 +436,8 @@ char *DeleteWebpageSpoolFile(URL *Url,int all)
  if(chdir(Url->proto))
    {err=PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return(err);}
 
- if(chdir(Url->host))
-   {err=PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return(err);}
+ if(chdir(Url->dir))
+   {err=PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);chdir("..");return(err);}
 
  /* Delete the file for the web page. */
 
@@ -439,25 +447,46 @@ char *DeleteWebpageSpoolFile(URL *Url,int all)
     DIR *dir=opendir(".");
 
     if(!dir)
-      {err=PrintMessage(Warning,"Cannot open current directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("../..");return(err);}
+      {err=PrintMessage(Warning,"Cannot open current directory '%s/%s' [%!s].",Url->proto,Url->dir);chdir("../..");return(err);}
 
     ent=readdir(dir);  /* skip .  */
     if(!ent)
-      {err=PrintMessage(Warning,"Cannot read current directory '%s/%s' [%!s].",Url->proto,Url->host);closedir(dir);chdir("../..");return(err);}
+      {err=PrintMessage(Warning,"Cannot read current directory '%s/%s' [%!s].",Url->proto,Url->dir);closedir(dir);chdir("../..");return(err);}
     ent=readdir(dir);  /* skip .. */
 
     while((ent=readdir(dir)))
       {
+       if(*ent->d_name=='D')
+         {
+          char *delurl;
+
+          if((delurl=FileNameToURL(ent->d_name)))
+            {
+             URL *delUrl=SplitURL(delurl);
+
+             chdir("../..");
+
+             if(ExistsLastTimeSpoolFile(delUrl))
+                DeleteLastTimeSpoolFile(delUrl);
+
+             chdir(Url->proto);
+             chdir(Url->dir);
+
+             free(delurl);
+             FreeURL(delUrl);
+            }
+         }
+
        if(unlink(ent->d_name))
-          err=PrintMessage(Warning,"Cannot unlink cached file '%s/%s/%s' [%!s].",Url->proto,Url->host,ent->d_name);
+          err=PrintMessage(Warning,"Cannot unlink cached file '%s/%s/%s' [%!s].",Url->proto,Url->dir,ent->d_name);
       }
 
     closedir(dir);
 
     chdir("..");
 
-    if(rmdir(Url->host))
-      err=PrintMessage(Warning,"Cannot delete what should be an empty directory '%s/%s' [%!s].",Url->proto,Url->host);
+    if(rmdir(Url->dir))
+      err=PrintMessage(Warning,"Cannot delete what should be an empty directory '%s/%s' [%!s].",Url->proto,Url->dir);
 
     chdir("..");
    }
@@ -468,13 +497,13 @@ char *DeleteWebpageSpoolFile(URL *Url,int all)
     int didstat=1;
 
     if(stat(".",&buf))
-       PrintMessage(Warning,"Cannot stat directory '%s/%s' [%!s].",Url->proto,Url->host);
+       PrintMessage(Warning,"Cannot stat directory '%s/%s' [%!s].",Url->proto,Url->dir);
     else
        didstat=1;
 
     *file='D';
     if(unlink(file))
-       err=PrintMessage(Warning,"Cannot unlink cached file '%s/%s/%s' [%!s].",Url->proto,Url->host,file);
+       err=PrintMessage(Warning,"Cannot unlink cached file '%s/%s/%s' [%!s].",Url->proto,Url->dir,file);
 
     *file='U';
     unlink(file);
@@ -491,6 +520,9 @@ char *DeleteWebpageSpoolFile(URL *Url,int all)
       }
 
     chdir("../..");
+
+    if(ExistsLastTimeSpoolFile(Url))
+       DeleteLastTimeSpoolFile(Url);
    }
 
  return(err);
@@ -501,9 +533,11 @@ char *DeleteWebpageSpoolFile(URL *Url,int all)
   Touch a file in a spool subdirectory.
 
   URL *Url The URL to touch.
+
+  time_t when The time to set the access time to.
   ++++++++++++++++++++++++++++++++++++++*/
 
-void TouchWebpageSpoolFile(URL *Url)
+void TouchWebpageSpoolFile(URL *Url,time_t when)
 {
  char *file;
 
@@ -512,15 +546,33 @@ void TouchWebpageSpoolFile(URL *Url)
  if(chdir(Url->proto))
    {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return;}
 
- if(chdir(Url->host))
-   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return;}
+ if(chdir(Url->dir))
+   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);chdir("..");return;}
 
  /* Touch the file for the web page. */
 
  file=URLToFileName(Url);
 
  *file='D';
- utime(file,NULL);
+
+ if(when)
+   {
+    struct stat buf;
+
+    if(stat(file,&buf))
+       utime(file,NULL);
+    else
+      {
+       struct utimbuf ubuf;
+
+       ubuf.actime=when;
+       ubuf.modtime=buf.st_mtime;
+
+       utime(file,&ubuf);
+      }
+   }
+ else
+    utime(file,NULL);
 
  /* Change dir back and tidy up. */
 
@@ -533,23 +585,22 @@ void TouchWebpageSpoolFile(URL *Url)
 /*++++++++++++++++++++++++++++++++++++++
   Check if a file in a spool subdirectory exists.
 
-  int ExistsWebpageSpoolFile Return a true value if the page exists.
+  time_t ExistsWebpageSpoolFile Return a the time the page was last accessed if the page exists.
 
   URL *Url The URL to check for.
   ++++++++++++++++++++++++++++++++++++++*/
 
-int ExistsWebpageSpoolFile(URL *Url)
+time_t ExistsWebpageSpoolFile(URL *Url)
 {
  struct stat buf;
  char *file;
- int exists=0;
 
  /* Change to the spool directory. */
 
  if(chdir(Url->proto))
     return(0);
 
- if(chdir(Url->host))
+ if(chdir(Url->dir))
    {chdir("..");return(0);}
 
  /* Stat the file for the web page. */
@@ -557,7 +608,8 @@ int ExistsWebpageSpoolFile(URL *Url)
  file=URLToFileName(Url);
  *file='D';
 
- exists=!stat(file,&buf);
+ if(stat(file,&buf))
+    buf.st_atime=0;
 
  /* Change dir back and tidy up. */
 
@@ -565,7 +617,7 @@ int ExistsWebpageSpoolFile(URL *Url)
 
  free(file);
 
- return(exists);
+ return(buf.st_atime);
 }
 
 
@@ -585,8 +637,8 @@ void CreateBackupWebpageSpoolFile(URL *Url)
  if(chdir(Url->proto))
    {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return;}
 
- if(chdir(Url->host))
-   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return;}
+ if(chdir(Url->dir))
+   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);chdir("..");return;}
 
  /* Create the filenames and rename the files. */
 
@@ -603,7 +655,7 @@ void CreateBackupWebpageSpoolFile(URL *Url)
  else
    {
     if(rename(orgfile,bakfile))
-       PrintMessage(Warning,"Cannot rename backup cached file '%s/%s/%s' to %s [%!s].",Url->proto,Url->host,bakfile,orgfile);
+       PrintMessage(Warning,"Cannot rename backup cached file '%s/%s/%s' to %s [%!s].",Url->proto,Url->dir,bakfile,orgfile);
 
     *bakfile=*orgfile='U';
     rename(orgfile,bakfile);
@@ -634,8 +686,8 @@ void RestoreBackupWebpageSpoolFile(URL *Url)
  if(chdir(Url->proto))
    {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return;}
 
- if(chdir(Url->host))
-   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return;}
+ if(chdir(Url->dir))
+   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);chdir("..");return;}
 
  /* Create the filenames and rename the files. */
 
@@ -650,7 +702,7 @@ void RestoreBackupWebpageSpoolFile(URL *Url)
  if(!stat(bakfile,&buf))
    {
     if(rename(bakfile,orgfile))
-       PrintMessage(Warning,"Cannot rename backup cached file '%s/%s/%s' to %s [%!s].",Url->proto,Url->host,bakfile,orgfile);
+       PrintMessage(Warning,"Cannot rename backup cached file '%s/%s/%s' to %s [%!s].",Url->proto,Url->dir,bakfile,orgfile);
 
     *bakfile=*orgfile='U';
     rename(bakfile,orgfile);
@@ -680,8 +732,8 @@ void DeleteBackupWebpageSpoolFile(URL *Url)
  if(chdir(Url->proto))
    {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return;}
 
- if(chdir(Url->host))
-   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return;}
+ if(chdir(Url->dir))
+   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);chdir("..");return;}
 
  /* Delete the file for the backup web page. */
 
@@ -691,7 +743,7 @@ void DeleteBackupWebpageSpoolFile(URL *Url)
  strcat(bakfile,"~");
 
  if(unlink(bakfile))
-    PrintMessage(Warning,"Cannot unlink backup cached file '%s/%s/%s' [%!s].",Url->proto,Url->host,bakfile);
+    PrintMessage(Warning,"Cannot unlink backup cached file '%s/%s/%s' [%!s].",Url->proto,Url->dir,bakfile);
 
  *bakfile='U';
  unlink(bakfile);
@@ -720,8 +772,8 @@ void CreateLockWebpageSpoolFile(URL *Url)
  if(chdir(Url->proto))
    {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return;}
 
- if(chdir(Url->host))
-   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return;}
+ if(chdir(Url->dir))
+   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);chdir("..");return;}
 
  /* Create the lock file for the web page. */
 
@@ -738,7 +790,7 @@ void CreateLockWebpageSpoolFile(URL *Url)
  else
    {
     if(link(orgfile,lokfile))
-       PrintMessage(Warning,"Cannot make a lock file for '%s/%s/%s' [%!s].",Url->proto,Url->host,orgfile);
+       PrintMessage(Warning,"Cannot make a lock file for '%s/%s/%s' [%!s].",Url->proto,Url->dir,orgfile);
    }
 
  /* Change dir back and tidy up. */
@@ -767,7 +819,7 @@ int ExistsLockWebpageSpoolFile(URL *Url)
  if(chdir(Url->proto))
     return(0);
 
- if(chdir(Url->host))
+ if(chdir(Url->dir))
    {chdir("..");return(0);}
 
  /* Stat the file for the web page. */
@@ -802,8 +854,8 @@ void DeleteLockWebpageSpoolFile(URL *Url)
  if(chdir(Url->proto))
    {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return;}
 
- if(chdir(Url->host))
-   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return;}
+ if(chdir(Url->dir))
+   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);chdir("..");return;}
 
  /* Delete the file for the backup web page. */
 
@@ -812,7 +864,7 @@ void DeleteLockWebpageSpoolFile(URL *Url)
  *lokfile='L';
 
  if(unlink(lokfile))
-    PrintMessage(Inform,"Cannot unlink lock file '%s/%s/%s' [%!s].",Url->proto,Url->host,lokfile);
+    PrintMessage(Inform,"Cannot unlink lock file '%s/%s/%s' [%!s].",Url->proto,Url->dir,lokfile);
 
  /* Change dir back and tidy up. */
 
@@ -822,128 +874,6 @@ void DeleteLockWebpageSpoolFile(URL *Url)
 }
 
 
-/*++++++++++++++++++++++++++++++++++++++
-  Delete a specified URL from the lasttime directory.
-
-  char *DeleteLastTimeSpoolFile Returns NULL if OK else error message.
-
-  URL *Url The URL to delete or NULL for all of them.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-char *DeleteLastTimeSpoolFile(URL *Url)
-{
- struct stat buf;
- char *err=NULL;
- char *name;
-
- /* Create the lasttime directory if needed and change to it */
-
- if(!Url)
-   {
-    int i;
-
-    if(lstat("lasttime",&buf))
-      {
-       PrintMessage(Inform,"Directory 'lasttime' does not exist [%!s]; creating one.");
-       if(mkdir("lasttime",0755))
-         {err=PrintMessage(Warning,"Cannot create directory 'lasttime' [%!s].");return(err);}
-      }
-    else
-       if(!S_ISDIR(buf.st_mode))
-         {err=PrintMessage(Warning,"The file 'lasttime' is not a directory.");return(err);}
-
-    for(i=1;i<=NUM_PREVTIME_DIR;i++)
-      {
-       char prevtime[16];
-
-       sprintf(prevtime,"prevtime%d",i);
-
-       if(lstat(prevtime,&buf))
-         {
-          PrintMessage(Inform,"Directory '%s' does not exist [%!s]; creating one.",prevtime);
-          if(mkdir(prevtime,0755))
-            {err=PrintMessage(Warning,"Cannot create directory '%s' [%!s].",prevtime);return(err);}
-         }
-       else
-          if(!S_ISDIR(buf.st_mode))
-            {err=PrintMessage(Warning,"The file '%s' is not a directory.",prevtime);return(err);}
-      }
-   }
-
- /* Delete the file for the request or all of them. */
-
- if(Url)
-   {
-    name=URLToFileName(Url);
-    *name='D';
-
-    if(chdir("lasttime"))
-      {err=PrintMessage(Warning,"Cannot change to directory 'lasttime' [%!s].");return(err);}
-
-    if(unlink(name))
-       err=PrintMessage(Warning,"Cannot unlink lasttime request 'lasttime/%s' [%!s].",name);
-
-    *name='U';
-    unlink(name);
-
-    free(name);
-
-    chdir("..");
-   }
- else
-   {
-    char prevtime[16],prevprevtime[16];
-    int i;
-
-    for(i=NUM_PREVTIME_DIR;i>=0;i--)
-      {
-       strcpy(prevprevtime,prevtime);
-
-       if(i)
-          sprintf(prevtime,"prevtime%d",i);
-       else
-          strcpy(prevtime,"lasttime");
-
-       if(i==NUM_PREVTIME_DIR)
-         {
-          DIR *dir;
-          struct dirent* ent;
-
-          if(chdir(prevtime))
-            {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",prevtime);continue;}
-
-          dir=opendir(".");
-
-          if(!dir)
-            {PrintMessage(Warning,"Cannot open current directory '%s' [%!s].",prevtime);chdir("..");continue;}
-
-          ent=readdir(dir);  /* skip .  */
-          if(!ent)
-            {PrintMessage(Warning,"Cannot read current directory '%s' [%!s].",prevtime);closedir(dir);chdir("..");continue;}
-          ent=readdir(dir);  /* skip .. */
-
-          while((ent=readdir(dir)))
-             if(unlink(ent->d_name))
-                PrintMessage(Warning,"Cannot unlink previous time page '%s/%s' [%!s].",prevtime,ent->d_name);
-
-          chdir("..");
-
-          if(rmdir(prevtime))
-             PrintMessage(Warning,"Cannot unlink previous time directory '%s' [%!s].",prevtime);
-         }
-       else
-          if(rename(prevtime,prevprevtime))
-             PrintMessage(Warning,"Cannot rename previous time directory '%s' to '%s' [%!s].",prevtime,prevprevtime);
-      }
-
-    if(mkdir("lasttime",0755))
-       err=PrintMessage(Warning,"Cannot create directory 'lasttime' [%!s].");
-   }
-
- return(err);
-}
-
-
 /*++++++++++++++++++++++++++++++++++++++
   Create a file in the lasttime directory.
 
@@ -958,6 +888,9 @@ int CreateLastTimeSpoolFile(URL *Url)
  char *file;
  int exists=0;
 
+ if(NoLasttimeIndex)
+    return(1);
+
  /* Change to the last time directory */
 
  if(chdir("lasttime"))
@@ -970,17 +903,17 @@ int CreateLastTimeSpoolFile(URL *Url)
 
  if(stat(file,&buf)==-1)
    {
-    char *name=(char*)malloc(strlen(Url->proto)+strlen(Url->host)+strlen(file)+8);
-    sprintf(name,"../%s/%s/%s",Url->proto,Url->host,file);
+    char *name=(char*)malloc(strlen(Url->proto)+strlen(Url->dir)+strlen(file)+8);
+    sprintf(name,"../%s/%s/%s",Url->proto,Url->dir,file);
 
-    if(symlink(name,file))
+    if(link(name,file))
       {PrintMessage(Warning,"Cannot create file 'lasttime/%s' [%!s].",file);}
     else
       {
        *file='U';
-       sprintf(name,"../%s/%s/%s",Url->proto,Url->host,file);
+       sprintf(name,"../%s/%s/%s",Url->proto,Url->dir,file);
 
-       if(symlink(name,file))
+       if(link(name,file))
          {PrintMessage(Warning,"Cannot create file 'lasttime/%s' [%!s].",file);}
       }
 
@@ -999,6 +932,164 @@ int CreateLastTimeSpoolFile(URL *Url)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Determine if a particular URL exists in the last time spool directory.
+
+  int ExistsLastTimeSpoolFile Returns 1 if the URL exists.
+
+  URL *Url The URL to look for.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+int ExistsLastTimeSpoolFile(URL *Url)
+{
+ struct stat buf;
+ int i,retval=0;
+ char *name=URLToFileName(Url);
+
+ *name='D';
+
+ for(i=0;i<=NUM_PREVTIME_DIR;i++)
+   {
+    char lasttime[40];
+
+    if(i)
+       sprintf(lasttime,"prevtime%d/%s",i,name);
+    else
+       sprintf(lasttime,"lasttime/%s",name);
+
+    if(!stat(lasttime,&buf))
+      {retval=1;break;}
+   }
+
+ free(name);
+
+ return(retval);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Delete a specified URL from the lasttime directory.
+
+  char *DeleteLastTimeSpoolFile Returns NULL if OK else error message.
+
+  URL *Url The URL to delete or NULL for all of them.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+char *DeleteLastTimeSpoolFile(URL *Url)
+{
+ struct stat buf;
+ char *err=NULL;
+ int i;
+ char *name=URLToFileName(Url);
+
+ for(i=0;i<=NUM_PREVTIME_DIR;i++)
+   {
+    char lasttime[16];
+
+    if(i)
+       sprintf(lasttime,"prevtime%d",i);
+    else
+       strcpy(lasttime,"lasttime");
+
+    /* Change to the last time directory */
+
+    if(chdir(lasttime))
+      {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",lasttime);}
+    else
+      {
+       *name='D';
+
+       if(!stat(name,&buf))
+         {
+          if(unlink(name))
+             err=PrintMessage(Warning,"Cannot unlink lasttime request '%s/%s' [%!s].",lasttime,name);
+
+          *name='U';
+          unlink(name);
+         }
+
+       chdir("..");
+      }
+   }
+
+ free(name);
+
+ return(err);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Cycle the URLs from the lasttime directory down to the prevtime directories.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void CycleLastTimeSpoolFile(void)
+{
+ char lasttime[16],prevlasttime[16];
+ struct stat buf;
+ int i;
+
+ /* Cycle the directories */
+
+ for(i=NUM_PREVTIME_DIR;i>=0;i--)
+   {
+    if(i)
+       sprintf(lasttime,"prevtime%d",i);
+    else
+       strcpy(lasttime,"lasttime");
+
+    /* Create it if it does not exist. */
+
+    if(lstat(lasttime,&buf))
+      {
+       PrintMessage(Inform,"Directory '%s' does not exist [%!s]; creating one.",lasttime);
+       if(mkdir(lasttime,DirPerm))
+          PrintMessage(Warning,"Cannot create directory '%s' [%!s].",lasttime);
+      }
+    else
+       if(!S_ISDIR(buf.st_mode))
+          PrintMessage(Warning,"The file '%s' is not a directory.",lasttime);
+
+    /* Delete the contents of the oldest one, rename the newer ones. */
+
+    if(i==NUM_PREVTIME_DIR)
+      {
+       DIR *dir;
+       struct dirent* ent;
+
+       if(chdir(lasttime))
+         {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",lasttime);continue;}
+
+       dir=opendir(".");
+
+       if(!dir)
+         {PrintMessage(Warning,"Cannot open current directory '%s' [%!s].",lasttime);chdir("..");continue;}
+
+       ent=readdir(dir);  /* skip .  */
+       if(!ent)
+         {PrintMessage(Warning,"Cannot read current directory '%s' [%!s].",lasttime);closedir(dir);chdir("..");continue;}
+       ent=readdir(dir);  /* skip .. */
+
+       while((ent=readdir(dir)))
+          if(unlink(ent->d_name))
+             PrintMessage(Warning,"Cannot unlink previous time page '%s/%s' [%!s].",lasttime,ent->d_name);
+
+       chdir("..");
+
+       if(rmdir(lasttime))
+          PrintMessage(Warning,"Cannot unlink previous time directory '%s' [%!s].",lasttime);
+      }
+    else
+       if(rename(lasttime,prevlasttime))
+          PrintMessage(Warning,"Cannot rename previous time directory '%s' to '%s' [%!s].",lasttime,prevlasttime);
+
+    strcpy(prevlasttime,lasttime);
+   }
+
+ if(mkdir("lasttime",DirPerm))
+    PrintMessage(Warning,"Cannot create directory 'lasttime' [%!s].");
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Open a file in the monitor directory to write into.
 
@@ -1026,7 +1117,7 @@ int CreateMonitorSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8],cha
  if(lstat("monitor",&buf))
    {
     PrintMessage(Inform,"Directory 'monitor' does not exist [%!s]; creating one.");
-    if(mkdir("monitor",0755))
+    if(mkdir("monitor",DirPerm))
       {PrintMessage(Warning,"Cannot create directory 'monitor' [%!s].");return(-1);}
    }
  else
@@ -1042,7 +1133,7 @@ int CreateMonitorSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8],cha
 
  *file='O';
 
- fd=open(file,O_WRONLY|O_CREAT|O_TRUNC,0644);
+ fd=open(file,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,FilePerm);
 
  if(fd==-1)
    {PrintMessage(Warning,"Cannot create file 'monitor/%s' [%!s].",file);}
@@ -1051,7 +1142,7 @@ int CreateMonitorSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8],cha
     int ufd,mfd;
 
     *file='U';
-    ufd=open(file,O_WRONLY|O_CREAT|O_TRUNC,0644);
+    ufd=open(file,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,FilePerm);
 
     if(ufd!=-1)
       {
@@ -1065,7 +1156,7 @@ int CreateMonitorSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8],cha
       }
 
     *file='M';
-    mfd=open(file,O_WRONLY|O_CREAT|O_TRUNC,0644);
+    mfd=open(file,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,FilePerm);
 
     if(mfd!=-1)
       {
@@ -1141,13 +1232,11 @@ long ReadMonitorTimesSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8]
    {
     int interval=7;
 
-    mfd=open(file,O_RDONLY);
+    mfd=open(file,O_RDONLY|O_BINARY);
     if(mfd)
       {
        char str[8];
-       if(read(mfd,str,8)<=0)
-          interval=7;
-       else
+       if(read(mfd,str,8)>0)
           interval=atoi(str);
        if(interval<0)
           interval=0;
@@ -1163,35 +1252,43 @@ long ReadMonitorTimesSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8]
     else
       {
        int i;
-       strcpy(DofM,"0000000000000000000000000000000"); for(i=0;i<31;i+=interval) DofM[i]='1';
+       if(interval)
+         {strcpy(DofM,"0000000000000000000000000000000"); for(i=0;i<31;i+=interval) DofM[i]='1';}
+       else
+          strcpy(DofM,"1111111111111111111111111111111");
        strcpy(DofW,"1111111");
       }
-    strcpy(HofD,"000000000000000000000000"); HofD[(mtime/3600)%24]='1';
-       {chdir("..");return(atime);}
+    if(interval)
+      {strcpy(HofD,"000000000000000000000000"); HofD[(mtime/3600)%24]='1';}
+    else
+       strcpy(HofD,"111111111111111111111111");
+
+    chdir("..");
+    return(atime);
    }
 
  /* Assume that the file is in the new format. */
 
- mfd=open(file,O_RDONLY);
+ mfd=open(file,O_RDONLY|O_BINARY);
  if(mfd)
    {
     char line[80];
+    line[78]=0;
 
-    if(read(mfd,line,80)>0)
-       if(sscanf(line,"%12s %31s %7s %24s",MofY,DofM,DofW,HofD)!=4 ||
-          strlen(MofY)!=12 || strlen(DofM)!=31 || strlen(DofW)!=7 || strlen(HofD)!=24)
-         {
-          strcpy(MofY,"111111111111");
-          strcpy(DofM,"1111111111111111111111111111111");
-          strcpy(DofW,"1111111");
-          strcpy(HofD,"100000000000000000000000");
-         }
+    if(read(mfd,line,80)!=78 ||
+       sscanf(line,"%12s %31s %7s %24s",MofY,DofM,DofW,HofD)!=4 ||
+       strlen(MofY)!=12 || strlen(DofM)!=31 || strlen(DofW)!=7 || strlen(HofD)!=24)
+      {
+       strcpy(MofY,"111111111111");
+       strcpy(DofM,"1111111111111111111111111111111");
+       strcpy(DofW,"1111111");
+       strcpy(HofD,"100000000000000000000000");
+      }
 
     close(mfd);
    }
 
  chdir("..");
-
  return(atime);
 }
 
@@ -1262,6 +1359,61 @@ char *DeleteMonitorSpoolFile(URL *Url)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Create a temporary spool file.
+
+  int CreateTempSpoolFile Returns the file descriptor.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+int CreateTempSpoolFile(void)
+{
+ char name[16];
+ struct stat buf;
+ int fd;
+
+ /* Create the outgoing directory if needed and change to it */
+
+ if(lstat("temp",&buf))
+   {
+    PrintMessage(Inform,"Directory 'temp' does not exist [%!s]; creating one.");
+    if(mkdir("temp",DirPerm))
+      {PrintMessage(Warning,"Cannot create directory 'temp' [%!s].");return(-1);}
+   }
+ else
+    if(!S_ISDIR(buf.st_mode))
+      {PrintMessage(Warning,"The file 'temp' is not a directory.");return(-1);}
+
+ /* Open the file */
+
+ sprintf(name,"temp/tmp.%ld",(long)getpid());
+
+ fd=open(name,O_RDWR|O_CREAT|O_TRUNC|O_BINARY,FilePerm);
+
+ if(fd==-1)
+    PrintMessage(Warning,"Cannot create temporary file '%s' [%!s].",name);
+
+ return(fd);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Close and delete the temporary spool file.
+
+  int fd The file descriptor.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void CloseTempSpoolFile(int fd)
+{
+ char name[16];
+
+ sprintf(name,"temp/tmp.%ld",(long)getpid());
+ if(unlink(name)==-1)
+    PrintMessage(Warning,"Cannot unlink temporary file '%s' [%!s].",name);
+
+ close(fd);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Convert a filename to a URL.
 
@@ -1285,7 +1437,7 @@ char *FileNameToURL(char *file)
 
  *copy='U';
 
- ufd=open(copy,O_RDONLY);
+ ufd=open(copy,O_RDONLY|O_BINARY);
 
  if(ufd==-1)
    {
