@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/wwwoffle.c 2.23 1998/11/30 20:20:41 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/wwwoffle.c 2.28 1999/08/15 14:40:14 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.5.
   A user level program to interact with the server.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97,98 Andrew M. Bishop
+  This file Copyright 1996,97,98,99 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -17,6 +17,7 @@
 #include <stdlib.h>
 #include <string.h>
 
+#include <limits.h>
 #include <unistd.h>
 #include <sys/stat.h>
 #include <fcntl.h>
@@ -37,20 +38,30 @@ static void usage(int verbose);
 typedef enum _Action
 {
  None,                          /*+ Undecided. +*/
+
  Online,                        /*+ Tell the server that we are online. +*/
  Autodial,                      /*+ Tell the server that we are in autodial mode. +*/
  Offline,                       /*+ Tell the server that we are offline. +*/
+
  Fetch,                         /*+ Tell the server to fetch the requested pages. +*/
+
  Config,                        /*+ Tell the server to re-read the configuration file. +*/
+
  Purge,                         /*+ Tell the server to purge pages. +*/
+
  Kill,                          /*+ Tell the server to exit. +*/
- Get,                           /*+ Tell the server to get pages. +*/
+
+ Get,                           /*+ Tell the server to GET pages. +*/
+ Post,                          /*+ Tell the server to POST a page. +*/
+ Put,                           /*+ Tell the server to PUT a page. +*/
+
  Output,                        /*+ Get a page from the server and output it. +*/
  OutputWithHeader               /*+ Get a page and the headers from the server and output it. +*/
 }
 Action;
 
 
+static void add_url_list(char **links);
 static void add_url_file(char *url_file);
 
 /*+ The list of URLs or files. +*/
@@ -105,6 +116,24 @@ int main(int argc, char** argv)
        continue;
       }
 
+    if(!strcmp(argv[i],"-post"))
+      {
+       if(action!=None)
+         {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
+       action=Post;
+       argv[i]=NULL;
+       continue;
+      }
+
+    if(!strcmp(argv[i],"-put"))
+      {
+       if(action!=None)
+         {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
+       action=Put;
+       argv[i]=NULL;
+       continue;
+      }
+
     if(!strcmp(argv[i],"-F"))
       {
        if(action!=None && action!=Get)
@@ -331,6 +360,12 @@ int main(int argc, char** argv)
  if(action==None)
     action=Get;
 
+ if((action==Post || action==Put) && n_url_file_list>1)
+   {
+    fprintf(stderr,"wwwoffle: The -post and -put options only allow one URL.\n\n");
+    usage(0);
+   }
+
  /* Initialise things. */
 
  if(!ConfigFile && !host && (env=getenv("WWWOFFLE_PROXY")))
@@ -376,12 +411,12 @@ int main(int argc, char** argv)
 
  /* The connections to the WWWOFFLE server. */
 
- if(action!=Get && action!=Output && action!=OutputWithHeader)
+ if(action!=Get && action!=Post && action!=Put && action!=Output && action!=OutputWithHeader)
    {
     int socket;
-    char *buffer=NULL;
+    char *line=NULL;
 
-    socket=OpenClientSocket(host?host:"localhost",port?port:WWWOFFLE_Port);
+    socket=OpenClientSocket(host?host:"localhost",port?port:WWWOFFLE_Port,ConnectTimeout);
     init_buffer(socket);
 
     if(socket==-1)
@@ -409,9 +444,9 @@ int main(int argc, char** argv)
     else
        write_string(socket,"WWWOFFLE BOGUS\r\n");
 
-    while((buffer=read_line(socket,buffer)))
+    while((line=read_line(socket,line)))
       {
-       fputs(buffer,stdout);
+       fputs(line,stdout);
        fflush(stdout);
       }
 
@@ -424,18 +459,19 @@ int main(int argc, char** argv)
    {
     URL *Url;
     struct stat buf;
-    char *refresh;
+    char *refresh=NULL;
 
-    refresh=CreateRefreshPath(recursive_depth,recursive_mode,force,
-                              stylesheets,images,frames,scripts,objects);
+    if(recursive_depth || recursive_mode || force || stylesheets || images || frames || scripts || objects)
+       refresh=CreateRefreshPath(recursive_depth,recursive_mode,force,
+                                 stylesheets,images,frames,scripts,objects);
 
     for(i=0;i<n_url_file_list;i++)
        if(strcmp(url_file_list[i],"-") && stat(url_file_list[i],&buf))
          {
           int socket;
-          char buffer256[257];
+          char buffer[READ_BUFFER_SIZE];
 
-          socket=OpenClientSocket(host?host:"localhost",port?port:HTTP_Port);
+          socket=OpenClientSocket(host?host:"localhost",port?port:HTTP_Port,ConnectTimeout);
           init_buffer(socket);
 
           if(socket==-1)
@@ -443,7 +479,10 @@ int main(int argc, char** argv)
 
           Url=SplitURL(url_file_list[i]);
 
-          printf("Getting: %s\n",Url->name);
+          if(refresh)
+             printf("Getting: %s (with recursive options).\n",Url->name);
+          else
+             printf("Getting: %s\n",Url->name);
 
           if(Url->user)
             {
@@ -452,24 +491,37 @@ int main(int argc, char** argv)
              sprintf(userpass1,"%s:%s",Url->user,Url->pass?Url->pass:"");
              userpass2=Base64Encode(userpass1,strlen(userpass1));
 
-             write_formatted(socket,"GET %s/?%s HTTP/1.0\r\n"
-                                    "Pragma: wwwoffle\r\n"
-                                    "Accept: */*\r\n"
-                                    "Authorization: Basic %s\r\n"
-                                    "\r\n",
-                             refresh,Url->name,userpass2);
+             if(refresh)
+                write_formatted(socket,"GET %s/?%s HTTP/1.0\r\n"
+                                       "Authorization: Basic %s\r\n"
+                                       "Accept: */*\r\n"
+                                       "\r\n",
+                                refresh,Url->name,userpass2);
+             else
+                write_formatted(socket,"GET %s HTTP/1.0\r\n"
+                                       "Authorization: Basic %s\r\n"
+                                       "Accept: */*\r\n"
+                                       "\r\n",
+                                Url->name,userpass2);
 
              free(userpass1);
              free(userpass2);
             }
           else
-             write_formatted(socket,"GET %s/?%s HTTP/1.0\r\n"
-                                    "Pragma: wwwoffle\r\n"
-                                    "Accept: */*\r\n"
-                                    "\r\n",
-                             refresh,Url->name);
+            {
+             if(refresh)
+                write_formatted(socket,"GET %s/?%s HTTP/1.0\r\n"
+                                       "Accept: */*\r\n"
+                                       "\r\n",
+                                refresh,Url->name);
+             else
+                write_formatted(socket,"GET %s HTTP/1.0\r\n"
+                                       "Accept: */*\r\n"
+                                       "\r\n",
+                                Url->name);
+            }
 
-          while(read_data(socket,buffer256,256)>0)
+          while(read_data(socket,buffer,READ_BUFFER_SIZE)>0)
              ;
 
           CloseSocket(socket);
@@ -481,7 +533,6 @@ int main(int argc, char** argv)
           int file;
           char *buffer=(char*)malloc(strlen(url_file_list[i])+256);
           char **links;
-          int j;
 
           if(strcmp(url_file_list[i],"-"))
             {
@@ -505,9 +556,9 @@ int main(int argc, char** argv)
              strcat(buffer,url_file_list[i]);
           else
             {
-             char cwd[256];
+             char cwd[PATH_MAX+1];
 
-             if(!getcwd(cwd,255))
+             if(!getcwd(cwd,PATH_MAX))
              cwd[0]=0;
 
              strcat(buffer,cwd);
@@ -518,57 +569,127 @@ int main(int argc, char** argv)
           Url=SplitURL(buffer);
           ParseHTML(file,Url);
 
+          if(stylesheets && (links=GetReferences(RefStyleSheet)))
+             add_url_list(links);
+
           if(images && (links=GetReferences(RefImage)))
-             for(j=0;links[j];j++)
-               {
-                URL *linkUrl=SplitURL(links[j]);
-
-                if(strcmp(linkUrl->proto,"file"))
-                   add_url_file(linkUrl->name);
-
-                FreeURL(linkUrl);
-               }
-
-          if((links=GetReferences(RefLink)))
-             for(j=0;links[j];j++)
-               {
-                URL *linkUrl=SplitURL(links[j]);
-
-                if(strcmp(linkUrl->proto,"file"))
-                   add_url_file(linkUrl->name);
-
-                FreeURL(linkUrl);
-               }
+             add_url_list(links);
 
           if(frames && (links=GetReferences(RefFrame)))
-             for(j=0;links[j];j++)
-               {
-                URL *linkUrl=SplitURL(links[j]);
+             add_url_list(links);
 
-                if(strcmp(linkUrl->proto,"file"))
-                   add_url_file(linkUrl->name);
+          if(scripts && (links=GetReferences(RefScript)))
+             add_url_list(links);
 
-                FreeURL(linkUrl);
-               }
+          if(objects && (links=GetReferences(RefObject)))
+             add_url_list(links);
+
+          if(objects && (links=GetReferences(RefInlineObject)))
+             add_url_list(links);
+
+          if((links=GetReferences(RefLink)))
+             add_url_list(links);
 
           FreeURL(Url);
+          free(buffer);
           if(file!=0)
              close(file);
          }
        else
           PrintMessage(Warning,"The file '%s' is not a regular file.",url_file_list[i]);
    }
+ else if(action==Post || action==Put)
+   {
+    URL *Url;
+    int socket,n,length=0;
+    char buffer[READ_BUFFER_SIZE];
+    char *data=(char*)malloc(READ_BUFFER_SIZE+1);
+
+    socket=OpenClientSocket(host?host:"localhost",port?port:HTTP_Port,ConnectTimeout);
+    init_buffer(socket);
+
+    if(socket==-1)
+       PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host?host:"localhost",port?port:HTTP_Port);
+
+    Url=SplitURL(url_file_list[0]);
+
+    if(action==Post)
+       printf("Posting: %s\n",Url->name);
+    else
+       printf("Putting: %s\n",Url->name);
+
+    while((n=read_data(0,data+length,READ_BUFFER_SIZE))>0)
+      {
+       length+=n;
+       data=(char*)realloc((void*)data,length+READ_BUFFER_SIZE+1);
+      }
+
+    if(Url->user)
+      {
+       char *userpass1=(char*)(malloc(strlen(Url->user)+(Url->pass?strlen(Url->pass):1)+3)),*userpass2;
+
+       sprintf(userpass1,"%s:%s",Url->user,Url->pass?Url->pass:"");
+       userpass2=Base64Encode(userpass1,strlen(userpass1));
+
+       if(action==Post)
+          write_formatted(socket,"POST %s HTTP/1.0\r\n"
+                                 "Authorization: Basic %s\r\n"
+                                 "Content-Type: application/x-www-form-urlencoded\r\n"
+                                 "Content-Length: %d\r\n"
+                                 "Accept: */*\r\n"
+                                 "\r\n",
+                          Url->name,userpass2,length);
+       else
+          write_formatted(socket,"PUT %s HTTP/1.0\r\n"
+                                 "Authorization: Basic %s\r\n"
+                                 "Content-Length: %d\r\n"
+                                 "Accept: */*\r\n"
+                                 "\r\n",
+                          Url->name,userpass2,length);
+
+       free(userpass1);
+       free(userpass2);
+      }
+    else
+      {
+       if(action==Post)
+          write_formatted(socket,"POST %s HTTP/1.0\r\n"
+                                 "Content-Type: application/x-www-form-urlencoded\r\n"
+                                 "Content-Length: %d\r\n"
+                                 "Accept: */*\r\n"
+                                 "\r\n",
+                          Url->name,length);
+       else
+          write_formatted(socket,"PUT %s HTTP/1.0\r\n"
+                                 "Content-Length: %d\r\n"
+                                 "Accept: */*\r\n"
+                                 "\r\n",
+                          Url->name,length);
+      }
+
+    write_data(socket,data,length);
+
+    write_data(socket,"\r\n",2);
+
+    while(read_data(socket,buffer,READ_BUFFER_SIZE)>0)
+       ;
+
+    CloseSocket(socket);
+
+    FreeURL(Url);
+    free(data);
+   }
  else /* action==Output or action==OutputWithHeader */
    {
     URL *Url;
     int socket;
-    char *buffer=NULL,buffer256[257];
+    char *line=NULL,buffer[READ_BUFFER_SIZE];
     int nbytes;
 
     if(!n_url_file_list)
        PrintMessage(Fatal,"No URL specified to output.");
 
-    socket=OpenClientSocket(host?host:"localhost",port?port:HTTP_Port);
+    socket=OpenClientSocket(host?host:"localhost",port?port:HTTP_Port,ConnectTimeout);
     init_buffer(socket);
 
     if(socket==-1)
@@ -603,14 +724,14 @@ int main(int argc, char** argv)
                               "\r\n",
                        Url->name);
 
-    if((buffer=read_line(socket,buffer)))
+    if((line=read_line(socket,line)))
       {
        char *willget="HTTP/1.0 404 WWWOFFLE Will Get\r\n";
        int status;
 
-       sscanf(buffer,"%*s %d",&status);
+       sscanf(line,"%*s %d",&status);
 
-       if(!strcmp(willget,buffer))
+       if(!strcmp(willget,line))
           fprintf(stderr,"The URL is not in the cache but has been requested.\n");
        else if((status>=300 && status<400) && action!=OutputWithHeader)
           fprintf(stderr,"The URL has been moved, check with a browser.\n");
@@ -619,18 +740,18 @@ int main(int argc, char** argv)
        else
          {
           if(action==OutputWithHeader)
-             fputs(buffer,stdout);
+             fputs(line,stdout);
 
-          while((buffer=read_line(socket,buffer)))
+          while((line=read_line(socket,line)))
             {
              if(action==OutputWithHeader)
-                fputs(buffer,stdout);
-             if(buffer[0]=='\r' || buffer[0]=='\n')
+                fputs(line,stdout);
+             if(line[0]=='\r' || line[0]=='\n')
                 break;
             }
 
-          while((nbytes=read_data(socket,buffer256,256))>0)
-             fwrite(buffer256,1,nbytes,stdout);
+          while((nbytes=read_data(socket,buffer,READ_BUFFER_SIZE))>0)
+             fwrite(buffer,1,nbytes,stdout);
          }
       }
     else
@@ -662,8 +783,8 @@ static void usage(int verbose)
 
  if(verbose)
     fprintf(stderr,
-            "(c) Andrew M. Bishop 1996,97,98 [       amb@gedanken.demon.co.uk ]\n"
-            "                                [http://www.gedanken.demon.co.uk/]\n"
+            "(c) Andrew M. Bishop 1996,97,98,99 [       amb@gedanken.demon.co.uk ]\n"
+            "                                   [http://www.gedanken.demon.co.uk/]\n"
             "\n");
 
  fprintf(stderr,
@@ -673,6 +794,7 @@ static void usage(int verbose)
          "       wwwoffle [-o|-O] <url>\n"
          "       wwwoffle [-g[Sisfo]] [-F] [-(d|r|R)[<depth>]] <url> ...\n"
          "       wwwoffle [-g[Sisfo]] [-F] [-(d|r|R)[<depth>]] [<file>|-] ...\n"
+         "       wwwoffle -post <url> | -put <url>\n"
          "\n"
          "Any of these can also take:  [-p <host>[:<port>] | -c <config-file>]\n"
          "The environment variable WWWOFFLE_PROXY can be set instead of -p or -c options.\n"
@@ -716,6 +838,12 @@ static void usage(int verbose)
             "                        (-r => URLs on the same host)\n"
             "                        (-R => URLs on any host)\n"
             "\n"
+            "wwwoffle -post <url> : Create a request using the POST method, the data is read\n"
+            "                       from stdin and appended to the request.  The user should\n"
+            "                       ensure that the data is correctly url-encoded.\n"
+            "wwwoffle -put <url>  : Create a request using the PUT method, the data is read\n"
+            "                       from stdin and appended to the request.\n"
+            "\n"
             " -p <host>[:<port>]  : The host name and port number to talk to the demon on.\n"
             "                       (Defaults to localhost for the server and\n"
             "                        %d for control port, %d for http proxy port).\n"
@@ -737,6 +865,28 @@ static void usage(int verbose)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Add a list of URLs to the list.
+
+  char **links The list of URLs to add.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void add_url_list(char **links)
+{
+ int i;
+
+ for(i=0;links[i];i++)
+   {
+    URL *linkUrl=SplitURL(links[i]);
+
+    if(strcmp(linkUrl->proto,"file"))
+       add_url_file(linkUrl->name);
+
+    FreeURL(linkUrl);
+   }
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Add a URL or a file to the list.
 
