@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/wwwoffle.h 2.55 1998/12/05 13:49:57 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/wwwoffle.h 2.65 1999/10/10 10:32:45 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.5a.
   A header file for all of the programs wwwoffle, wwwoffled.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97,98 Andrew M. Bishop
+  This file Copyright 1996,97,98,99 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -44,10 +44,22 @@
 
 /*+ The number of pretime directories to create (lasttime history). +*/
 #define NUM_PREVTIME_DIR 3
+/* If this is changed then to see the links to them you will need to edit
+   /var/spool/wwwoffle/html/messages/IndexLastTime-Tail.html. */
+
+/*+ The size of the buffer to use when reading from the cache or a socket. +*/
+#define READ_BUFFER_SIZE 1024
+
+/*+ The default permission for creating directories. +*/
+#define DEF_DIR_PERM  0755
+/*+ The default permission for creating files. +*/
+#define DEF_FILE_PERM 0644
 
 /* The values in the section above can be changed if required, see INSTALL.
    ------------------------------------------------------------ */
 
+#include <time.h>
+
 #include "misc.h"
 
 /* In connect.c */
@@ -69,8 +81,8 @@ char *DeleteOutgoingSpoolFile(URL *Url);
 
 int OpenWebpageSpoolFile(int rw,URL *Url);
 char *DeleteWebpageSpoolFile(URL *Url,int all);
-void TouchWebpageSpoolFile(URL *Url);
-int ExistsWebpageSpoolFile(URL *Url);
+void TouchWebpageSpoolFile(URL *Url,time_t when);
+time_t ExistsWebpageSpoolFile(URL *Url);
 
 void CreateBackupWebpageSpoolFile(URL *Url);
 void RestoreBackupWebpageSpoolFile(URL *Url);
@@ -80,38 +92,53 @@ void CreateLockWebpageSpoolFile(URL *Url);
 void DeleteLockWebpageSpoolFile(URL *Url);
 int ExistsLockWebpageSpoolFile(URL *Url);
 
-char *DeleteLastTimeSpoolFile(URL *Url);
 int CreateLastTimeSpoolFile(URL *Url);
+char *DeleteLastTimeSpoolFile(URL *Url);
+int ExistsLastTimeSpoolFile(URL *Url);
+void CycleLastTimeSpoolFile(void);
 
 int CreateMonitorSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8],char HofD[25]);
 long ReadMonitorTimesSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8],char HofD[25]);
 char *DeleteMonitorSpoolFile(URL *Url);
 
+int CreateTempSpoolFile(void);
+void CloseTempSpoolFile(int fd);
+
 char *FileNameToURL(char *file);
 char *URLToFileName(URL *Url);
 
 /* In parse.c */
 
-char *ParseRequest(int fd,char **request_head,char **request_body);
+char *ParseRequest(int fd,Header **request_head,Body **request_body);
 
-int RequestChanges(int fd,char **request_head);
-char *MovedLocation(URL *Url,char *reply_head);
-char *RequestURL(URL *Url,char *referer);
+int RequestChanges(int fd,Header *request_head);
+char *MovedLocation(URL *Url,Header *reply_head);
+Header *RequestURL(URL *Url,char *referer);
 
-char *ModifyRequest(URL *Url,char *request_head);
+void ModifyRequest(URL *Url,Header *request_head);
 
-char *MakeRequestAuthorised(char *proxy,char *request_head);
-char *MakeRequestNonProxy(char *request_head);
+void MakeRequestAuthorised(char *proxy,Header *request_head);
+void MakeRequestNonProxy(Header *request_head);
 
-int ParseReply(int fd,URL *Url,char **reply_head);
+int ParseReply(int fd,URL *Url,Header **reply_head);
 
 int SpooledPageStatus(URL *Url);
 
-char *GetHTTPHeader(char *head,char* line);
+void ModifyReply(Header *reply_head);
+
+Header *CreateHeader(char *top,int type);
+void AddToHeader(Header *head,char *key,char *val);
+void RemoveFromHeader(Header *head,char* key,char *val);
+char *GetHeader(Header *head,char* key,char *val);
+char *HeaderString(Header *head);
+void FreeHeader(Header *head);
+
+Body *CreateBody(int length);
+void FreeBody(Body *body);
 
 /* In messages.c (messages.l) */
 
-void LocalPage(int fd,char *path,char *request_head);
+void LocalPage(int fd,char *path,Header *request_head);
 char *HTMLMessage(int fd,int status_val,char *status_str,char *location,char *template, ...);
 char *HTMLMessageHead(int fd,int status_val,char *status_str, ...);
 char *HTMLMessageBody(int fd,char *template, ...);
@@ -122,24 +149,25 @@ void IndexPage(int fd,URL *Url);
 
 /* In control.c */
 
-void ControlPage(int fd,URL *Url,char *request_head,char *request_body);
+void ControlPage(int fd,URL *Url,Header *request_head,Body *request_body);
 
 /* In configedit.c */
 
-void ConfigEditPage(int fd,char *args,char *request_body);
+void ConfigEditPage(int fd,char *args,Body *request_body);
 
 /* In refresh.c */
 
-char *RefreshPage(int fd,URL *Url,char *request_body,int *recurse);
+char *RefreshPage(int fd,URL *Url,Body *request_body,int *recurse);
 void ParseRecurseOptions(char *method);
 int RecurseFetch(URL *Url,int new);
 int RecurseFetchRelocation(URL *Url,char *location);
 char *CreateRefreshPath(int recursive_depth,int recursive_mode,int force,
                         int stylesheets,int images,int frames,int scripts,int objects);
+int RefreshForced(void);
 
 /* In monitor.c */
 
-void MonitorPage(int fd,URL *Url,char *request_body);
+void MonitorPage(int fd,URL *Url,Body *request_body);
 void RequestMonitoredPages(void);
 void MonitorTimes(URL *Url,int *last,int *next);
 
@@ -149,6 +177,6 @@ int wwwoffles(int online,int browser,int client);
 
 /* In htdig.c */
 
-void HTDigPage(int fd,URL *Url,char *head);
+void HTDigPage(int fd,URL *Url,Header *head);
 
 #endif /* WWWOFFLE_H */
