@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/wwwoffled.c 2.14 1998/11/30 20:20:41 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/wwwoffled.c 2.18 1999/08/08 09:22:33 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.5.
   A demon program to maintain the database and spawn the servers.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97,98 Andrew M. Bishop
+  This file Copyright 1996,97,98,99 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -62,6 +62,9 @@ int n_servers=0,                /*+ in total. +*/
 /*+ The wwwoffle client file descriptor when fetching. +*/
 int fetch_fd=-1;
 
+/*+ The pids of the servers. +*/
+int server_pids[MAX_SERVERS];
+
 /*+ The pids of the servers that are fetching. +*/
 int fetch_pids[MAX_FETCH_SERVERS];
 
@@ -80,6 +83,9 @@ static sig_atomic_t got_sigchld=0;
 /*+ True if run as a demon +*/
 static int detached=1;
 
+/*+ True if the pid of the daemon should be printed at startup +*/
+static int print_pid=0;
+
 
 /*++++++++++++++++++++++++++++++++++++++
   The main program.
@@ -110,6 +116,12 @@ int main(int argc, char** argv)
        continue;
       }
 
+    if(!strcmp(argv[i],"-p"))
+      {
+       print_pid=1;
+       continue;
+      }
+
     if(!strcmp(argv[i],"-c"))
       {
        if(++i>=argc)
@@ -127,6 +139,9 @@ int main(int argc, char** argv)
  for(i=0;i<MAX_FETCH_SERVERS;i++)
     fetch_pids[i]=0;
 
+ for(i=0;i<MAX_SERVERS;i++)
+    server_pids[i]=0;
+
  InitErrorHandler("wwwoffled",1,!detached);
 
  if(ReadConfigFile(2))
@@ -162,9 +177,11 @@ int main(int argc, char** argv)
  if(wwwoffle_fd==-1)
     PrintMessage(Fatal,"Cannot create WWWOFFLE server socket.");
 
+ umask(0);
+
  if(stat(SpoolDir,&buf))
    {
-    err=mkdir(SpoolDir,0755);
+    err=mkdir(SpoolDir,DirPerm);
     if(err==-1 && errno!=EEXIST)
        PrintMessage(Fatal,"Cannot create spool directory %s [%!s].",SpoolDir);
     stat(SpoolDir,&buf);
@@ -281,6 +298,7 @@ int main(int argc, char** argv)
     if(got_sigchld)
       {
        int pid, status;
+       int isserver=0;
 
        /* To avoid race conditions, reset the flag before fetching the status */
        got_sigchld=0;
@@ -295,12 +313,20 @@ int main(int argc, char** argv)
          else if(WIFSIGNALED(status))
             exitval=-WTERMSIG(status);
 
-         if(exitval>=0)
-            PrintMessage(Inform,"Child wwwoffles exited with status %d (pid=%d).",exitval,pid);
-         else
-            PrintMessage(Important,"Child wwwoffles terminated by signal %d (pid=%d).",-exitval,pid);
+         for(i=0;i<MaxServers;i++)
+            if(server_pids[i]==pid)
+              {
+               n_servers--;
+               server_pids[i]=0;
+               isserver=1;
 
-         n_servers--;
+               if(exitval>=0)
+                  PrintMessage(Inform,"Child wwwoffles exited with status %d (pid=%d).",exitval,pid);
+               else
+                  PrintMessage(Important,"Child wwwoffles terminated by signal %d (pid=%d).",-exitval,pid);
+
+               break;
+              }
 
          /* Check if the child that terminated is one of the fetching wwwoffles */
          for(i=0;i<MaxFetchServers;i++)
@@ -321,7 +347,8 @@ int main(int argc, char** argv)
             fetching=0;
         }
 
-        PrintMessage(Debug,"Currently running: %d servers total, %d fetchers.",n_servers,n_fetch_servers);
+       if(isserver)
+          PrintMessage(Debug,"Currently running: %d servers total, %d fetchers.",n_servers,n_fetch_servers);
       }
 
     /* The select timed out or we got a signal. If we are currently fetching,
@@ -347,12 +374,13 @@ int main(int argc, char** argv)
  CloseSocket(wwwoffle_fd);
 
  if(n_servers)
-    PrintMessage(Important,"Exit signalled - waiting for children.");
+    PrintMessage(Important,"Exit signalled - waiting for %d child wwwoffles servers.",n_servers);
  else
     PrintMessage(Important,"Exit signalled.");
 
  while(n_servers)
    {
+    int i;
     int pid,status,exitval=0;
 
     while((pid=waitpid(-1,&status,0))>0)
@@ -362,12 +390,19 @@ int main(int argc, char** argv)
        else if(WIFSIGNALED(status))
           exitval=-WTERMSIG(status);
 
-       if(exitval>=0)
-          PrintMessage(Inform,"Child wwwoffles exited with status %d (pid=%d).",exitval,pid);
-       else
-          PrintMessage(Important,"Child wwwoffles terminated by signal %d (pid=%d).",-exitval,pid);
+       for(i=0;i<MaxServers;i++)
+          if(server_pids[i]==pid)
+            {
+             n_servers--;
+             server_pids[i]=0;
 
-       n_servers--;
+             if(exitval>=0)
+                PrintMessage(Inform,"Child wwwoffles exited with status %d (pid=%d).",exitval,pid);
+             else
+                PrintMessage(Important,"Child wwwoffles terminated by signal %d (pid=%d).",-exitval,pid);
+
+             break;
+            }
       }
    }
 
@@ -409,6 +444,8 @@ static void usage(int verbose)
             "\n"
             "   -d [<log-level>]    : Do not detach from the terminal and use stderr.\n"
             "                         0 <= log-level <= %d (default in config file).\n"
+            "\n"
+            "   -p                  : Print the pid of wwwoffled on stdout (unless -d).\n"
             "\n",ConfigFile,Fatal+1);
 
  exit(0);
@@ -435,7 +472,13 @@ static void demoninit(void)
  if(pid==-1)
     PrintMessage(Fatal,"Cannot fork() to detach(2) [%!s].");
  else if(pid)
+   {
+    /* print the child pid on stdout as requested */
+    if(print_pid)
+       printf("%d\n", pid);
+
     exit(0);
+   }
 
  /* Close fds */
  close(STDIN_FILENO);
